#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : arithmetic.py
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-05-14 11:42
   Description : description what the main function of this file
"""
from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedPairBase, SharedTensorBase
import numpy as np
from stensorflow.exception.exception import StfValueException, StfEqualException, StfCondException
import tensorflow as tf
from typing import Union
from stensorflow.global_var import StfConfig
from stensorflow.basic.protocol.bilinear_map import BM_PrivateTensor_SharedPair, BM_PrivateTensor_PrivateTensor, \
    BM_SharedPair_SharedPair, BM_SharedPair_PrivateTensor
from stensorflow.homo_enc.homo_mat_mul import matmul_homo_offline, matmul_homo_online, matmul_homo
from stensorflow.homo_enc.homo_mul import mul_homo, mul_homo_offline, mul_homo_online


def add(x: Union[PrivateTensorBase, SharedPairBase, int, float, np.ndarray, tf.Tensor],
        y: Union[PrivateTensorBase, SharedPairBase, int, float, np.ndarray, tf.Tensor]) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    x_is_private = isinstance(x, PrivateTensorBase)
    x_is_pair = isinstance(x, SharedPairBase)
    x_is_public = isinstance(x, (np.ndarray, tf.Tensor, float, int))
    y_is_private = isinstance(y, PrivateTensorBase)
    y_is_pair = isinstance(y, SharedPairBase)
    y_is_public = isinstance(y, (np.ndarray, tf.Tensor, float, int))

    if x_is_private and y_is_private:
        if x.owner == y.owner:
            return PrivateTensorBase.__add__(x, y)
        else:
            fixedpoint = min(x.fixedpoint, y.fixedpoint)
            altered_x = x.dup_with_precision(fixedpoint)
            altered_y = y.dup_with_precision(fixedpoint)
            return SharedPairBase(ownerL=altered_x.owner, ownerR=altered_y.owner, xL=altered_x.to_SharedTensor(),
                                  xR=altered_y.to_SharedTensor(), fixedpoint=fixedpoint)
    elif x_is_private and y_is_pair:
        fixedpoint = min(x.fixedpoint, y.fixedpoint)
        altered_x = x.dup_with_precision(fixedpoint)
        altered_y = y.dup_with_precision(fixedpoint)
        if x.owner == y.ownerL:
            with tf.device(y.ownerL):
                zL = altered_x.to_SharedTensor() + altered_y.xL
            return SharedPairBase(ownerL=y.ownerL, ownerR=y.ownerR, xL=zL, xR=altered_y.xR, fixedpoint=fixedpoint)
        elif x.owner == y.ownerR:
            with tf.device(y.ownerR):
                zR = altered_x.to_SharedTensor() + altered_y.xR
            return SharedPairBase(ownerL=y.ownerL, ownerR=y.ownerR, xL=altered_y.xL, xR=zR, fixedpoint=fixedpoint)
        else:
            raise Exception("must have x.owner==y.ownerL or x.owner==y.ownerR")
    elif x_is_private and y_is_public:
        y1 = PrivateTensorBase(owner=x.owner, fixedpoint=x.fixedpoint)
        y1.load_from_tf_tensor(y)
        # return x + y1
        return add(x, y1)

    elif x_is_pair and y_is_private:
        return add(y, x)
    elif x_is_pair and y_is_pair:
        fixedpoint = min(x.fixedpoint, y.fixedpoint)
        altered_x = x.dup_with_precision(fixedpoint)
        altered_y = y.dup_with_precision(fixedpoint)
        if x.ownerL != y.ownerL or x.ownerR != y.ownerR:
            y = y.mirror()
        if x.ownerL != y.ownerL or x.ownerR != y.ownerR:
            raise Exception("must have x.ownerL=y.ownerL and x.ownerR==y.ownerR")
        with tf.device(x.ownerL):
            zL = altered_x.xL + altered_y.xL
        with tf.device(x.ownerR):
            zR = altered_x.xR + altered_y.xR
        return SharedPairBase(ownerL=y.ownerL, ownerR=y.ownerR, xL=zL, xR=zR, fixedpoint=fixedpoint)
    elif x_is_pair and y_is_public:
        y1 = PrivateTensorBase(owner=x.ownerR, fixedpoint=x.fixedpoint)
        y1.load_from_tf_tensor(y)
        return add(x, y1)
    elif x_is_public and y_is_private:
        return add(y, x)
    elif x_is_public and y_is_pair:
        return add(y, x)
    else:
        raise NotImplementedError


def sub(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase]) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    return add(x, -y)


def mul3p(x: Union[PrivateTensorBase, SharedPairBase],
          y: Union[PrivateTensorBase, SharedPairBase, int, float, np.ndarray, tf.Tensor], fixed_point=None,
          prf_flag=None) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    if fixed_point is not None:
        pass
    elif not hasattr(y, "fixedpoint"):
        fixed_point = x.fixedpoint
    else:
        fixed_point = max(x.fixedpoint, y.fixedpoint)

    if prf_flag is None:
        prf_flag = StfConfig.prf_flag

    x_is_private = isinstance(x, PrivateTensorBase)
    x_is_pair = isinstance(x, SharedPairBase)
    y_is_private = isinstance(y, PrivateTensorBase)
    y_is_pair = isinstance(y, SharedPairBase)
    y_is_public = isinstance(y, (np.ndarray, tf.Tensor, float, int))

    if x_is_private and y_is_private:
        if x.owner == y.owner:
            result = PrivateTensorBase.__mul__(x, y)
        else:
            result = BM_PrivateTensor_PrivateTensor(x, y, lambda _x, _y: SharedTensorBase.__mul__(_x, _y),
                                                    prf_flag=False)
    elif x_is_private and y_is_pair:
        result = BM_PrivateTensor_SharedPair(x, y, lambda _x, _y: SharedTensorBase.__mul__(_x, _y), prf_flag=prf_flag)
    elif x_is_pair and y_is_private:
        result = BM_SharedPair_PrivateTensor(x, y, lambda _x, _y: SharedTensorBase.__mul__(_x, _y), prf_flag=prf_flag)
    elif x_is_pair and y_is_pair:
        result = BM_SharedPair_SharedPair(x, y, lambda _x, _y: SharedTensorBase.__mul__(_x, _y), prf_flag=prf_flag)
    elif (x_is_private or x_is_pair) and y_is_public:
        result = rmul(y, x)
    else:
        raise NotImplementedError
    return result.dup_with_precision(fixed_point)


def mul2p(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase],
          fixed_point=None) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    # print("matmul2p")
    x_is_private = isinstance(x, PrivateTensorBase)
    x_is_shared = isinstance(x, SharedPairBase)
    y_is_private = isinstance(y, PrivateTensorBase)
    y_is_shared = isinstance(y, SharedPairBase)

    if not x_is_private and not y_is_private:
        raise NotImplementedError
    if StfConfig.pre_produce_flag:
        # print("StfConfig.pre_produce_flag=", StfConfig.pre_produce_flag)
        if x_is_private and y_is_private:
            x_adjoint = x.random_uniform_adjoint()
            y_adjoint = y.random_uniform_adjoint()
        elif x_is_private and y_is_shared:
            x_adjoint = x.random_uniform_adjoint()
            y_owner = y.ownerR if y.ownerL == x.owner else y.ownerL
            y_adjoint = PrivateTensorBase(owner=y_owner, fixedpoint=y.fixedpoint,
                                          inner_value=y.xL.random_uniform_adjoint().inner_value)
        elif x_is_shared and y_is_private:
            y_adjoint = y.random_uniform_adjoint()
            x_owner = x.ownerL if y.owner == x.ownerR else x.ownerR
            x_adjoint = PrivateTensorBase(owner=x_owner, fixedpoint=x.fixedpoint,
                                          inner_value=x.xL.random_uniform_adjoint().inner_value)
        else:
            x_adjoint = x.random_uniform_adjoint()
            y_adjoint = y.random_uniform_adjoint()
        z_adjoint = mul_homo_offline(x_adjoint, y_adjoint)
        if StfConfig.offline_model:
            pass
        else:
            x_adjoint.unserialize()
            y_adjoint.unserialize()
            z_adjoint.unserialize()
        result = mul_homo_online(x, y, x_adjoint, y_adjoint, z_adjoint, fixed_point=fixed_point)
    else:
        result = mul_homo(x, y, fixed_point=fixed_point)

    return result


def mul(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase],
        fixed_point=None, prf_flag=None) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    if StfConfig.parties == 2:
        return mul2p(x, y, fixed_point)
    else:
        return mul3p(x, y, fixed_point, prf_flag)


def matmul3p(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase],
             fixed_point=None, prf_flag=None) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    if fixed_point is None:
        fixed_point = max(x.fixedpoint, y.fixedpoint)
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag

    x_is_private = isinstance(x, PrivateTensorBase)
    x_is_pair = isinstance(x, SharedPairBase)
    y_is_private = isinstance(y, PrivateTensorBase)
    y_is_pair = isinstance(y, SharedPairBase)

    if x_is_private and y_is_private:
        if x.owner == y.owner:
            result = PrivateTensorBase.__matmul__(x, y)
        else:
            result = BM_PrivateTensor_PrivateTensor(x, y, lambda _x, _y: SharedTensorBase.__matmul__(_x, _y),
                                                    prf_flag=prf_flag)
    elif x_is_private and y_is_pair:
        result = BM_PrivateTensor_SharedPair(x, y, lambda _x, _y: SharedTensorBase.__matmul__(_x, _y),
                                             prf_flag=prf_flag)
    elif x_is_pair and y_is_private:
        result = BM_SharedPair_PrivateTensor(x, y, lambda _x, _y: SharedTensorBase.__matmul__(_x, _y),
                                             prf_flag=prf_flag)
    elif x_is_pair and y_is_pair:
        result = BM_SharedPair_SharedPair(x, y, lambda _x, _y: SharedTensorBase.__matmul__(_x, _y), prf_flag=prf_flag)
    else:
        raise Exception("type exception for type(x)={}, type(y)={}".format(type(x), type(y)))
    return result.dup_with_precision(fixed_point)


def matmul2p(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase],
             fixed_point=None) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    x_is_private = isinstance(x, PrivateTensorBase)
    x_is_shared = isinstance(x, SharedPairBase)
    y_is_private = isinstance(y, PrivateTensorBase)
    y_is_shared = isinstance(y, SharedPairBase)

    if not x_is_private and not y_is_private:
        raise NotImplementedError
    if StfConfig.pre_produce_flag:
        if x_is_private and y_is_private:
            x_adjoint = x.random_uniform_adjoint()
            y_adjoint = y.random_uniform_adjoint()
        elif x_is_private and y_is_shared:
            x_adjoint = x.random_uniform_adjoint()
            y_owner = y.ownerR if y.ownerL == x.owner else y.ownerL
            y_adjoint = PrivateTensorBase(owner=y_owner, fixedpoint=y.fixedpoint,
                                          inner_value=y.xL.random_uniform_adjoint().inner_value)
        elif x_is_shared and y_is_private:
            y_adjoint = y.random_uniform_adjoint()
            x_owner = x.ownerL if y.owner == x.ownerR else x.ownerR
            x_adjoint = PrivateTensorBase(owner=x_owner, fixedpoint=x.fixedpoint,
                                          inner_value=x.xL.random_uniform_adjoint().inner_value)
        else:
            x_adjoint = x.random_uniform_adjoint()
            y_adjoint = y.random_uniform_adjoint()

        z_adjoint = matmul_homo_offline(x_adjoint, y_adjoint)
        if StfConfig.offline_model:
            pass
        else:
            x_adjoint.unserialize()
            y_adjoint.unserialize()
            z_adjoint.unserialize()
        result = matmul_homo_online(x, y, x_adjoint, y_adjoint, z_adjoint, fixed_point=fixed_point)
    else:
        result = matmul_homo(x, y, fixed_point=fixed_point)

    return result


def matmul(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase],
           fixed_point=None, prf_flag=None) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    if StfConfig.parties == 2:
        return matmul2p(x, y, fixed_point)
    else:
        return matmul3p(x, y, fixed_point, prf_flag)


def rmul_matmul(x: Union[int, float, np.ndarray, tf.Tensor], y: Union[PrivateTensorBase, SharedPairBase],
                mul_func, fixed_point=None) -> Union[PrivateTensorBase, SharedPairBase]:
    if isinstance(x, int) or isinstance(x, np.ndarray) and x.dtype == np.int \
            or isinstance(x, tf.Tensor) and x.dtype in (
            tf.int8, tf.int16, tf.int32, tf.int64, tf.uint8, tf.uint16, tf.uint32, tf.uint64):
        fixedpoint_x = 0
    else:
        fixedpoint_x = StfConfig.default_fixed_point
        x = tf.cast(x * (1 << fixedpoint_x), 'int64')

    if isinstance(y, PrivateTensorBase):
        with tf.device(y.owner):
            inner_value = tf.cast(mul_func(x, y.inner_value), 'int64')
            if y.module is not None:
                inner_value = inner_value % y.module
        w = PrivateTensorBase(owner=y.owner, inner_value=inner_value, module=y.module,
                              fixedpoint=fixedpoint_x + y.fixedpoint)
    elif isinstance(y, SharedPairBase):
        x = SharedTensorBase(inner_value=x, module=y.xL.module)
        with tf.device(y.ownerL):
            yL = mul_func(x, y.xL)

        with tf.device(y.ownerR):
            yR = mul_func(x, y.xR)
        w = SharedPairBase(ownerL=y.ownerL, ownerR=y.ownerR, xL=yL, xR=yR, fixedpoint=fixedpoint_x + y.fixedpoint)
    else:
        raise Exception("y must be PrivateTensorBase or SharedPairBase")
    if fixed_point is None:
        fixed_point = max(fixedpoint_x, y.fixedpoint)
        return w.dup_with_precision(new_fixedpoint=fixed_point)


def rmul(x: Union[int, float, np.ndarray, tf.Tensor], y: Union[PrivateTensorBase, SharedPairBase],
         fixed_point=None) -> Union[PrivateTensorBase, SharedPairBase]:
    return rmul_matmul(x, y, lambda a, b: a * b, fixed_point)


#
def rmatmul(x: Union[np.ndarray, tf.Tensor], y: Union[PrivateTensorBase, SharedPairBase], fixed_point=None) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    return rmul_matmul(x, y, lambda a, b: a @ b, fixed_point)


def truediv(x: Union[PrivateTensorBase, SharedPairBase], y: Union[int, float, np.ndarray, tf.Tensor]) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    return rmul(1.0 / y, x)



#!/usr/bin/env python
# coding=utf-8
"""
   Alipay.com Inc.
   Copyright (c) 2004-2021 All Rights Reserved.
   ------------------------------------------------------
   File Name : bilinear_triangle
   Author : qizhi.zqz
   Email: qizhi.zqz@alibaba-inc.com
   Create Time : 2021/11/30 下午2:40
   Description : description what the main function of this file
"""

from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedPairBase, SharedTensorBase
from stensorflow.exception.exception import StfTypeException
from stensorflow.basic.basic_class.private import PrivateTensor
from stensorflow.basic.basic_class.pair import SharedPair
import tensorflow as tf
from stensorflow.global_var import StfConfig
from stensorflow.random.random import get_seed
from typing import Union

"""
(x~,y~)->u
(y~,z~)->v
(z~,x~)->w
"""


def BT_forward_SharedPair_SharedPair(x: SharedPairBase, y: SharedPairBase, f_xy, f_yz, f_zx, prf_flag=None):
    # f: (SharedTensorBase, SharedTensorBase)->SharedTensorBase

    if prf_flag is None:
        prf_flag = StfConfig.prf_flag

    if x.xL.module != y.xL.module:
        raise Exception("must have x.xL.module==y.xL.module")
    if x.ownerL != y.ownerL or x.ownerR != y.ownerR:
        y = y.mirror()
    if x.ownerL != y.ownerL or x.ownerR != y.ownerR:
        raise Exception("must have x.ownerL==y.ownerL and x.ownerR==y.ownerR")
    with tf.device(StfConfig.RS[0]):
        if prf_flag:
            seed_xL = get_seed()
            seed_xR = get_seed()
            seed_yL = get_seed()
            seed_yR = get_seed()
            seed_zL = get_seed()
            seed_zR = get_seed()
            seed_u = get_seed()
            seed_v = get_seed()
            seed_w = get_seed()
        else:
            seed_xL = None
            seed_xR = None
            seed_yL = None
            seed_yR = None
            seed_zL = None
            seed_zR = None
            seed_u = None
            seed_v = None
            seed_w = None

        xL_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_xL)
        xR_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_xR)
        yL_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed_yL)
        yR_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed_yR)
        u = f_xy(xL_adjoint + xR_adjoint, yL_adjoint + yR_adjoint)  # SharedTensor
        zL_adjoint = u.random_uniform_adjoint(seed_zL)
        zR_adjoint = u.random_uniform_adjoint(seed_zR)
        v = f_yz(yL_adjoint + yR_adjoint, zL_adjoint + zR_adjoint)  # SharedTensor
        w = f_zx(zL_adjoint + zR_adjoint, xL_adjoint + xR_adjoint)

        uL = u.random_uniform_adjoint(seed_u)
        uR = u - uL
        vL = v.random_uniform_adjoint(seed_v)
        vR = v - vL
        wL = w.random_uniform_adjoint(seed_w)
        wR = w - wL

    with tf.device(x.ownerL):
        delta_xL = x.xL - xL_adjoint
        delta_yL = y.xL - yL_adjoint
    with tf.device(y.ownerR):
        delta_xR = x.xR - xR_adjoint
        delta_yR = y.xR - yR_adjoint
    with tf.device(x.ownerL):
        delta_x_onL = delta_xL + delta_xR
        delta_y_onL = delta_yL + delta_yR
        rL = f_xy(delta_x_onL, y.xL) + f_xy(xL_adjoint, delta_y_onL) + uL
    with tf.device(x.ownerR):
        delta_x_onR = delta_xL + delta_xR
        delta_y_onR = delta_yL + delta_yR
        rR = f_xy(delta_x_onR, y.xR) + f_xy(xR_adjoint, delta_y_onR) + uR
    r = SharedPairBase(xL=rL, xR=rR, ownerL=x.ownerL, ownerR=x.ownerR, fixedpoint=x.fixedpoint + y.fixedpoint)
    if prf_flag:
        s = (delta_x_onL, delta_x_onR, delta_y_onL, delta_y_onR,
             seed_xL, seed_xR, seed_yL, seed_yR, seed_zL, seed_zR,
             seed_v, vR, seed_w, wR, x.fixedpoint, y.fixedpoint)
    else:
        s = (delta_x_onL, delta_x_onR, delta_y_onL, delta_y_onR,
            xL_adjoint, xR_adjoint, yL_adjoint, yR_adjoint, zL_adjoint, zR_adjoint,
            vL, vR, wL, wR, x.fixedpoint, y.fixedpoint)
    return r, s


def BT_backward_SharedPair_SharedPair(s, z: SharedPairBase, f_yz, f_zx, prf_flag):
    if prf_flag:
        delta_x_onL, delta_x_onR, delta_y_onL, delta_y_onR, \
        seed_xL, seed_xR, seed_yL, seed_yR, seed_zL, seed_zR, \
        seed_v, vR, seed_w, wR, x_fixedpoint, y_fixedpoint = s
        with tf.device(z.ownerL):
            xL_adjoint = delta_x_onL.random_uniform_adjoint(seed_xL)
            yL_adjoint = delta_y_onL.random_uniform_adjoint(seed_yL)
            zL_adjoint = z.to_SharedTensor_like().random_uniform_adjoint(seed_zL)
            vL = vR.random_uniform_adjoint(seed_v)
            wL = wR.random_uniform_adjoint(seed_w)
        with tf.device(z.ownerR):
            xR_adjoint = delta_x_onR.random_uniform_adjoint(seed_xR)
            yR_adjoint = delta_y_onR.random_uniform_adjoint(seed_yR)
            zR_adjoint = z.to_SharedTensor_like().random_uniform_adjoint(seed_zR)
    else:
        delta_x_onL, delta_x_onR, delta_y_onL, delta_y_onR, \
        xL_adjoint, xR_adjoint, yL_adjoint, yR_adjoint, zL_adjoint, zR_adjoint, \
        vL, vR, wL, wR, x_fixedpoint, y_fixedpoint = s
    with tf.device(z.ownerL):
        delta_zL = z.xL - zL_adjoint
    with tf.device(z.ownerR):
        delta_zR = z.xR - zR_adjoint
    with tf.device(z.ownerL):
        delta_z_onL = delta_zL+delta_zR
        vL = f_yz(delta_y_onL, z.xL) + f_yz(yL_adjoint, delta_z_onL) + vL
        wL = f_zx(z.xL, delta_x_onL) + f_zx(delta_z_onL, xL_adjoint) + wL
    with tf.device(z.ownerR):
        delta_z_onR = delta_zL+delta_zR
        vR = f_yz(delta_y_onR, z.xR) + f_yz(yR_adjoint, delta_z_onR) + vR
        wR = f_zx(z.xR, delta_x_onR) + f_zx(delta_z_onR, xR_adjoint) + wR
    v = SharedPairBase(xL=vL, xR=vR, ownerL=z.ownerL, ownerR=z.ownerR, fixedpoint=y_fixedpoint + z.fixedpoint)
    w = SharedPairBase(xL=wL, xR=wR, ownerL=z.ownerL, ownerR=z.ownerR, fixedpoint=z.fixedpoint + x_fixedpoint)
    return v, w



def BT_forward_PrivateTensor_SharedPair(x: PrivateTensorBase, y: SharedPairBase, f_xy, f_yz, f_zx, prf_flag=None):
    # f: (PrivateTensorBase, SharedTensorBase)->SharedTensorBase

    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if x.module != y.xL.module:
        raise Exception("must have x.module==y.module")
    if y.ownerL == y.ownerR:
        raise Exception("must have y.ownerL!=y.ownerR")
    if x.owner == y.ownerL:
        with tf.device(StfConfig.RS[0]):
            if prf_flag:
                seed_x = get_seed()
                seed_yL = get_seed()
                seed_yR = get_seed()
                seed_zL = get_seed()
                seed_zR = get_seed()
                seed_u = get_seed()
                seed_v = get_seed()
                seed_w = get_seed()
            else:
                seed_x = None
                seed_yL = None
                seed_yR = None
                seed_zL = None
                seed_zR = None
                seed_u = None
                seed_v = None
                seed_w = None
            x_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_x)
            yL_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed_yL)
            yR_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed_yR)
            u = f_xy(x_adjoint, yL_adjoint + yR_adjoint)  # SharedTensor
            zL_adjoint = u.random_uniform_adjoint(seed_zL)
            zR_adjoint = u.random_uniform_adjoint(seed_zR)
            v = f_yz(yL_adjoint + yR_adjoint, zL_adjoint + zR_adjoint)  # SharedTensor
            w = f_zx(zL_adjoint + zR_adjoint, x_adjoint)

            uL = u.random_uniform_adjoint(seed_u)
            uR = u - uL
            vL = v.random_uniform_adjoint(seed_v)
            vR = v - vL
            wL = w.random_uniform_adjoint(seed_w)
            wR = w - wL

        with tf.device(x.owner):
            if prf_flag:
                x_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_x)
                yL_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed_yL)
            delta_x = x.to_SharedTensor() - x_adjoint
            delta_yL = y.xL - yL_adjoint
        with tf.device(y.ownerR):
            if prf_flag:
                yR_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed_yR)
            delta_yR = y.xR - yR_adjoint
        with tf.device(x.owner):
            delta_y_onL = delta_yL + delta_yR
            rL = f_xy(delta_x, y.xL) + f_xy(x_adjoint, delta_y_onL) + uL
        with tf.device(y.ownerR):
            delta_y_onR = delta_yL + delta_yR
            rR = f_xy(delta_x, y.xR) + uR
        r = SharedPairBase(xL=rL, xR=rR, ownerL=x.owner, ownerR=y.ownerR, fixedpoint=x.fixedpoint + y.fixedpoint)
        if prf_flag:
            s = (delta_x, delta_y_onL, delta_y_onR,
                 seed_x, seed_yL, seed_yR, seed_zL, seed_zR,
                 seed_v, vR, seed_w, wR, x.fixedpoint, y.fixedpoint)
        else:
            s = (delta_x, delta_y_onL, delta_y_onR,
                x_adjoint, yL_adjoint, yR_adjoint, zL_adjoint, zR_adjoint,
                vL, vR, wL, wR, x.fixedpoint, y.fixedpoint)
    elif x.owner == y.ownerR:
        y = y.mirror()
        r, s = BT_forward_PrivateTensor_SharedPair(x, y, f_xy, f_yz, f_zx, prf_flag=prf_flag)
    else:
        raise Exception("must have x.owner==y.ownerL or x.owner==y.ownerR")
    return r, s




def BT_backward_PrivateTensor_SharedPair(s, z: SharedPairBase, f_yz, f_zx, x: PrivateTensorBase, prf_flag=None):

    if prf_flag:
        delta_x, delta_y_onL, delta_y_onR, \
         seed_x, seed_yL, seed_yR, seed_zL, seed_zR, \
         seed_v, vR, seed_w, wR, x_fixedpoint, y_fixedpoint = s
        with tf.device(z.ownerL):
            x_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_x)
            yL_adjoint = delta_y_onL.random_uniform_adjoint(seed_yL)
            zL_adjoint = z.to_SharedTensor_like().random_uniform_adjoint(seed_zL)
            vL = vR.random_uniform_adjoint(seed_v)
            wL = wR.random_uniform_adjoint(seed_w)
        with tf.device(z.ownerR):
            yR_adjoint = delta_y_onR.random_uniform_adjoint(seed_yR)
            zR_adjoint = z.to_SharedTensor_like().random_uniform_adjoint(seed_zR)
    else:
        delta_x, delta_y_onL, delta_y_onR, \
        x_adjoint, yL_adjoint, yR_adjoint, zL_adjoint, zR_adjoint, \
        vL, vR, wL, wR, x_fixedpoint, y_fixedpoint = s
    with tf.device(z.ownerL):
        delta_zL = z.xL - zL_adjoint
    with tf.device(z.ownerR):
        delta_zR = z.xR - zR_adjoint
    with tf.device(z.ownerL):
        delta_z_onL = delta_zL+delta_zR
        vL = f_yz(delta_y_onL, z.xL) + f_yz(yL_adjoint, delta_z_onL) + vL
        if x.owner == z.ownerL:
            wL = f_zx(z.xL, delta_x) + f_zx(delta_z_onL, x_adjoint) + wL
        elif x.owner == z.ownerR:
            wL = f_zx(z.xL, delta_x) + wL
        else:
            raise Exception("must have x.owner == z.ownerL or x.owner == z.ownerR")
    with tf.device(z.ownerR):
        delta_z_onR = delta_zL+delta_zR
        vR = f_yz(delta_y_onR, z.xR) + f_yz(yR_adjoint, delta_z_onR) + vR
        if x.owner == z.ownerL:
            wR = f_zx(z.xR, delta_x) + wR
        elif x.owner == z.ownerR:
            wR = f_zx(z.xR, delta_x) + f_zx(delta_z_onR, x_adjoint) + wR
        else:
            raise Exception("must have x.owner == z.ownerL or x.owner == z.ownerR")
    v = SharedPairBase(xL=vL, xR=vR, ownerL=z.ownerL, ownerR=z.ownerR, fixedpoint=y_fixedpoint + z.fixedpoint)
    w = SharedPairBase(xL=wL, xR=wR, ownerL=z.ownerL, ownerR=z.ownerR, fixedpoint=z.fixedpoint + x_fixedpoint)
    return v, w



def BT_forward_SharedPair_PrivateTensor(x: SharedPairBase, y: PrivateTensorBase, f_xy, f_yz, f_zx, prf_flag=None):
    x1 = y
    y1 = x
    g_X1Y1 = lambda a, b: f_xy(b, a)
    g_Y1Z = lambda a, b: f_zx(b, a)
    g_ZX1 = lambda a, b: f_yz(b, a)
    return BT_forward_PrivateTensor_SharedPair(x1, y1, g_X1Y1, g_Y1Z, g_ZX1, prf_flag)


def BT_backward_SharedPair_PrivateTensor(s, z: SharedPairBase, f_yz, f_zx, y: PrivateTensorBase, prf_flag=None):
    x1 = y
    g_Y1Z = lambda a, b: f_zx(b, a)
    g_ZX1 = lambda a, b: f_yz(b, a)
    u, v = BT_backward_PrivateTensor_SharedPair(s, z, g_Y1Z, g_ZX1, x1, prf_flag)
    return v, u

class BiliinearTriangle():
    """
    (x,y)->u
    (y,z)->v
    (z,x)->w
    """
    def __init__(self, f_xy, f_yz, f_zx):
        self.f_xy = f_xy
        self.f_yz = f_yz
        self.f_zx = f_zx
        self.s = None
        self.x = None
        self.y = None
        self.prf_flag = StfConfig.prf_flag

    def compute_u(self, x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase]):
        self.x = x
        self.y = y
        is_private_x = isinstance(x, PrivateTensorBase)
        is_private_y = isinstance(y, PrivateTensorBase)
        if is_private_x and is_private_y:
            raise NotImplementedError
        elif is_private_x and not is_private_y:
            r, s = BT_forward_PrivateTensor_SharedPair(x, y, self.f_xy, self.f_yz, self.f_zx, self.prf_flag)
        elif not is_private_x and is_private_y:
            r, s = BT_forward_SharedPair_PrivateTensor(x, y, self.f_xy, self.f_yz, self.f_zx, self.prf_flag)
        else:
            r, s = BT_forward_SharedPair_SharedPair(x, y, self.f_xy, self.f_yz, self.f_zx, self.prf_flag)
        self.s = s
        return SharedPair.from_SharedPairBase(r)

    def compute_vw(self, z: SharedPairBase):
        is_private_x = isinstance(self.x, PrivateTensorBase)
        is_private_y = isinstance(self.y, PrivateTensorBase)

        if is_private_x and is_private_y:
            raise NotImplementedError
        elif is_private_x and not is_private_y:
            u, v = BT_backward_PrivateTensor_SharedPair(self.s, z, self.f_yz, self.f_zx, self.x,self.prf_flag)
        elif not is_private_x and is_private_y:
            u, v = BT_backward_SharedPair_PrivateTensor(self.s, z, self.f_yz, self.f_zx, self.y, self.prf_flag)
        else:
            u, v = BT_backward_SharedPair_SharedPair(self.s, z, self.f_yz, self.f_zx, self.prf_flag)
        return SharedPair.from_SharedPairBase(u), SharedPair.from_SharedPairBase(v)




#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : float_compare
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-09-24 19:54
   Description : description what the main function of this file
   export PYTHONPATH=$PYTHONPATH:/Users/qizhi.zqz/projects/morse-stf/morse-stf
"""
import numpy as np
import pandas as pd
import argparse
import tensorflow as tf
from stensorflow.basic.basic_class.private import PrivateTensor
from stensorflow.basic.basic_class.base import SharedTensorBase, SharedPairBase, get_device
from stensorflow.basic.protocol.compare import geq_SharedTensorBase_SharedTensorBase, \
    greater_SharedTensorBase_SharedTensorBase, less_SharedTensorBase_SharedTensorBase, \
    leq_SharedTensorBase_SharedTensorBase
from typing import Union
import time as time

def to_bits(x: tf.Tensor, little_endian=True) -> tf.Tensor:
    # if x.dtype != tf.int64:
    #     raise Exception("must have x.dtype==tf.int64")
    y = tf.bitcast(x, 'uint8')
    z = tf.bitwise.bitwise_and(tf.expand_dims(y, axis=-1), [1, 2, 4, 8, 16, 32, 64, 128])
    w = tf.bitwise.right_shift(z, [0, 1, 2, 3, 4, 5, 6, 7])
    new_shape = y.shape.as_list()
    new_shape[-1] = -1
    u = tf.reshape(w, new_shape)
    # if order == "decrease":
    if not little_endian:
        u = tf.reverse(u, axis=[-1])
    return tf.cast(u, 'int64')


def f2bits_keep_order(x: tf.Tensor) -> tf.Tensor:
    y = to_bits(x, little_endian=False)
    y_shape = y.shape.as_list()
    y = tf.reshape(y, [-1, y_shape[-1]])
    sign = y[:, 0:1]
    y = (y + sign) % 2
    z = tf.concat([1 - sign, y[:, 1:]], axis=1)
    return tf.reshape(z, y_shape)


def compare_float_float(x_owner, x: tf.Tensor, y_owner, y: tf.Tensor, relation: str) -> SharedPairBase:
    x_owner=get_device(x_owner)
    y_owner=get_device(y_owner)
    
    with tf.device(x_owner):
        x2b = f2bits_keep_order(x)
        
        x2b = SharedTensorBase(inner_value=x2b, module=2)
    with tf.device(y_owner):
        y2b = f2bits_keep_order(y)
        
        y2b = SharedTensorBase(inner_value=y2b, module=2)
    if relation == "less":
        z = less_SharedTensorBase_SharedTensorBase(x2b, y2b, x_owner=x_owner, y_owner=y_owner)
    elif relation == "leq":
        z = leq_SharedTensorBase_SharedTensorBase(x2b, y2b, x_owner=x_owner, y_owner=y_owner)
    elif relation == "greater":
        z = greater_SharedTensorBase_SharedTensorBase(x2b, y2b, x_owner=x_owner, y_owner=y_owner)
    elif relation == "geq":
        z = geq_SharedTensorBase_SharedTensorBase(x2b, y2b, x_owner=x_owner, y_owner=y_owner)
    else:
        raise Exception("relation must be less, leq, greater or geq")
    return z


def compare_int_int(x_owner, x: tf.Tensor, y_owner, y: tf.Tensor, relation: str) -> SharedPairBase:
    x_owner=get_device(x_owner)
    y_owner=get_device(y_owner)
    with tf.device(x_owner):
        x2b = to_bits(x, little_endian=False)
        
        x2b = SharedTensorBase(inner_value=x2b, module=2)
    with tf.device(y_owner):
        y2b = to_bits(y, little_endian=False)
        
        y2b = SharedTensorBase(inner_value=y2b, module=2)



    if relation == "less":
        z = less_SharedTensorBase_SharedTensorBase(x2b, y2b, x_owner=x_owner, y_owner=y_owner)
    elif relation == "leq":
        z = leq_SharedTensorBase_SharedTensorBase(x2b, y2b, x_owner=x_owner, y_owner=y_owner)
    elif relation == "greater":
        z = greater_SharedTensorBase_SharedTensorBase(x2b, y2b, x_owner=x_owner, y_owner=y_owner)
    elif relation == "geq":
        z = geq_SharedTensorBase_SharedTensorBase(x2b, y2b, x_owner=x_owner, y_owner=y_owner)
    else:
        raise Exception("relation must be less, leq, greater or geq")
    return z


#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : softmax
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-12-01 14:17
   Description : description what the main function of this file
"""

from stensorflow.basic.basic_class.pair import SharedPair
import tensorflow as tf
from stensorflow.exception.exception import StfTypeException
from stensorflow.basic.operator.relu import relu
from stensorflow.basic.operator.sigmoid import sin2pi
import numpy as np
from stensorflow.global_var import StfConfig
from stensorflow.basic.basic_class.private import PrivateTensor
from typing import Union


def softmax_bak(x: Union[SharedPair, PrivateTensor]):
    if isinstance(x, SharedPair):
        y = relu(x) # + x.ones_like()
        z = y.reduce_sum(axis=1, keepdims=True)
        # return ~ (sin2pi(z, T=2**41)*(2**40/np.pi)) * y
        return ~sin2pi(z, T=2 ** 14) * y / (2 ** 13 / np.pi)
    elif isinstance(x, PrivateTensor):
        with tf.device(x.owner):
            y = tf.nn.softmax(x.to_tf_tensor())
            y = tf.cast(y * (2 ** x.fixedpoint), 'int64')
            z = PrivateTensor(owner=x.owner, fixedpoint=x.fixedpoint,
                              inner_value=y, module=x.module, op_map=x.op_map)
            return z
    else:
        raise StfTypeException("x", "SharedPair or PrivateTensor", type(x))



def softmax(x: Union[SharedPair, PrivateTensor]):
    if isinstance(x, SharedPair):
        y = x.ones_like() / x.shape[-1]
        for _ in range(StfConfig.softmax_iter_num):
            # formula of Qizhi Zhang
            # y = y + (x - (y * x).reduce_sum(axis=-1, keepdims=True)) * y / StfConfig.softmax_iter_num
            y = y + (x - (y.expend_dims(axis=[-2]) @ x.expend_dims(axis=-1)).squeeze(axis=-1)) * y / StfConfig.softmax_iter_num

        return y
    elif isinstance(x, PrivateTensor):
        with tf.device(x.owner):
            y = tf.nn.softmax(x.to_tf_tensor())
            y = tf.cast(y * (2 ** x.fixedpoint), 'int64')
            z = PrivateTensor(owner=x.owner, fixedpoint=x.fixedpoint,
                              inner_value=y, module=x.module, op_map=x.op_map)
            return z
    else:
        raise StfTypeException("x", "SharedPair or PrivateTensor", type(x))



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : base
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-05-14 19:58
   Description : description what the main function of this file
"""

import tensorflow as tf
import tensorflow.python as python
from stensorflow.global_var import StfConfig
from stensorflow.exception.exception import StfEqualWarning, StfCondException, StfEqualException, StfTypeException, \
    StfDTypeException, StfValueException, StfNoneException, StfException
import numpy as np
from stensorflow.random.random import gen_rint64, gen_rint64_from_seed
from typing import Union
import random
import os
import string

random.seed(0)


def cycle_lshift_tensor(x: tf.Tensor, index: tf.Tensor):
    """
    Cycle left shift x by index.
    :param x: tf.Tensor
    :param index: tf.Tensor where x.shape[:-1]==index.shape
    :return: y: tf.Tensor, where the vector y[i1,..., in-1,:] is the
            cycle left shift of x[i1,..., in-1,:] by index[i1, ...in-1] for any i1, ..., in-1
    """
    if x.shape[:-1] != index.shape:
        raise StfEqualException(var1_name="x.shape[:-1]", var2_name="index.shape",
                                var1=x.shape[:-1], var2=index.shape)
    module = x.shape[-1]
    index = index % module
    one_to_n = tf.constant(list(range(module)), dtype='int64')
    index_ext = (tf.expand_dims(index, axis=-1) + tf.reshape(one_to_n, shape=[1] * len(index.shape) + [module]))
    index_ext = index_ext % one_to_n.shape[0]

    b = tf.gather(params=x, indices=index_ext, axis=len(index.shape),
                  batch_dims=len(index.shape))
    return b


def cycle_rshift_tensor(x: tf.Tensor, index: tf.Tensor):
    """
    Cycle right shift x by index
    :param x: tf.Tensor
    :param index: tf.Tensor where x.shape[:-1]==index.shape
    :return: y: tf.Tensor, where the vector y[i1,..., in-1,:] is the
            cycle right shift of x[i1,..., in-1,:] by index[i1, ...in-1] for any i1, ..., in-1
    """
    return cycle_lshift_tensor(x, -index)


class SharedTensorBase:
    """
    inner_value:  Tensorflow Tensor or Variable of dtype int64 orint32
    module: int
    shape: list
    must have inner_value is not None or shape is not None
    """

    def __init__(self, inner_value=None, module: int = None, shape=None):

        self.module = module
        self.inner_value = inner_value
        if self.inner_value is not None:
            if self.module is not None:
                inner_value %= self.module
            if not isinstance(inner_value, tf.Tensor):
                self.inner_value = tf.constant(inner_value, dtype='int64')
            else:
                self.inner_value = inner_value
            if self.inner_value.dtype != tf.dtypes.int64 and self.inner_value.dtype != tf.dtypes.int32:
                raise StfDTypeException(obj_name="self.inner_value", expect_dtypes=[tf.dtypes.int32, tf.dtypes.int64],
                                        real_dtype=self.inner_value.dtype)

            self.shape = self.inner_value.shape.as_list()
        elif shape is not None:
            self.shape = shape
        else:
            raise StfNoneException(obj_name="inner_value or shape")
        self.name = ''.join(random.sample(string.ascii_letters + string.digits, 16))

    def __repr__(self) -> str:
        return 'SharedTensorBase(module={}, shape={})'.format(self.module, self.shape)

    @property
    def device(self):
        return self.inner_value.device

    def check_inner_value_is_not_None(self):
        if self.inner_value is None:
            raise StfNoneException(obj_name="self.inner_value")

    def check_module_equal(self, other):
        if self.module != other.module:
            raise StfEqualException("self.module", "other.module", self.module, other.module)

    def __getitem__(self, item):
        self.check_inner_value_is_not_None()
        inner_value = self.inner_value[item]
        z = SharedTensorBase(inner_value=inner_value, module=self.module)
        return z

    def random_uniform_adjoint(self, seed=None):
        """

        :param seed: seed for generate random number
        :return:     generate  pseudo-random SharedTensorBase with same shape and
                    module with self
        """
        if seed is not None:
            x = gen_rint64_from_seed(shape=self.shape, seed=seed)
        else:
            x = gen_rint64(self.shape)
            # x = tf.random.uniform(shape=self.shape, minval=-(1 << 63), maxval=(1 << 63) - 1, dtype='int64')

        adjoint = SharedTensorBase(inner_value=x, module=self.module)
        return adjoint

    def random_stateless_uniform_adjoint(self):
        """

        :param seed: seed for generate random number
        :return:     generate  pseudo-random SharedTensorBase with same shape and
                    module with self
        """
        if self.module is not None:
            x = tf.random.stateless_uniform(shape=self.shape, seed=[0, 0], minval=0, maxval=self.module, dtype='int64')
        else:
            x = tf.random.stateless_uniform(shape=self.shape, seed=[0, 0], minval=None, maxval=None, dtype='int64')
        adjoint = SharedTensorBase(inner_value=x, module=self.module)
        return adjoint

    def split(self, size_splits, axis=0, num=None):
        """
        Splits self into a list of sub tensors.

        :param size_splits: Either an integer indicating the number of splits along
          `axis` or a 1-D integer `Tensor` or Python list containing the sizes of
          each output tensor along `axis`. If a scalar, then it must evenly divide
          `value.shape[axis]`; otherwise the sum of sizes along the split axis
          must match that of the `self`.
        :param axis: An integer or scalar `int32` `Tensor`. The dimension along which to
        split. Must be in the range `[-rank(value), rank(value))`. Defaults to 0.
        :param num: Optional, used to specify the number of outputs when it cannot be
        inferred from the shape of `size_splits`.
        :return:
        """
        self.check_inner_value_is_not_None()
        inner_values = tf.split(self.inner_value, size_splits, axis, num)
        vs = tuple(map(lambda inner_value: SharedTensorBase(inner_value=inner_value, module=self.module), inner_values))
        return vs

    def slice(self, begin, size):
        """
        The usage is refer to tf.slice
        :param begin: An `int32` or `int64` `Tensor`.
        :param size: An `int32` or `int64` `Tensor`.
        :return: A `SharedTensorBase` of the same shape and module as self
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.slice(self.inner_value, begin=begin, size=size)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def to_compress_tensor(self, dtype: tf.dtypes.DType = tf.int64) -> tf.Tensor:
        """
        Compress self to a tf.Tensor of dtype
        :param dtype:  a tf.dtypes.DType object
        :return:  a tf.Tensor of dtype
        """
        self.check_inner_value_is_not_None()
        if self.module is None:
            return tf.bitcast(self.inner_value, dtype)
        else:
            # x = tf.reshape(self.inner_value % self.module, [-1])
            x = tf.reshape(self.inner_value, [-1])
            size = x.shape.as_list()[0]
            capacity = int((8 * dtype.size - 1) // np.log2(self.module))  # how much log(module) in dtype
            new_size = int(np.ceil(1.0 * size / capacity))
            x = tf.pad(x, paddings=[[0, capacity * new_size - size]])
            x = tf.reshape(x, [new_size, capacity])
            y = tf.cast(x, dtype) * tf.constant(np.power(self.module, np.arange(capacity)), dtype=dtype,
                                                shape=[1, capacity])
            return tf.reduce_sum(y, axis=[1])

    def decompress_from(self, y: tf.Tensor, shape=None):
        """
        Decompress from tf.Tensor
        :param y: tf.Tensor
        :param shape:  self.shape
        """
        if shape is not None:
            if shape != self.shape:
                raise Exception(
                    "must have shape is None or shape==self.shape, but shape={}, self.shape={}".format(shape,
                                                                                                       self.shape))
        else:
            shape = self.shape

        if self.module is None:
            self.inner_value = tf.bitcast(y, 'int64')
        else:
            capacity = int((8 * y.dtype.size - 1) // np.log2(self.module))  # how much log(module) in dtype
            div_num = tf.constant(np.power(self.module, np.arange(capacity)), dtype=y.dtype, shape=[1, capacity])
            x = (tf.reshape(y, [-1, 1]) // div_num) % self.module
            x = tf.cast(x, 'int64')
            x = tf.reshape(x, [-1])
            size = int(np.prod(shape))
            x = tf.slice(x, [0], [size])
            self.inner_value = tf.reshape(x, shape)

    def decompress_from_to_new(self, y: tf.Tensor, shape=None):
        """
        Decompress from tf.Tensor
        :param y: tf.Tensor
        :param shape:  self.shape
        """
        if shape is not None:
            if shape != self.shape:
                raise Exception(
                    "must have shape is None or shape==self.shape, but shape={}, self.shape={}".format(shape,
                                                                                                       self.shape))
        else:
            shape = self.shape

        if self.module is None:
            inner_value = tf.bitcast(y, 'int64')
        else:
            capacity = int((8 * y.dtype.size - 1) // np.log2(self.module))  # how much log(module) in dtype
            div_num = tf.constant(np.power(self.module, np.arange(capacity)), dtype=y.dtype, shape=[1, capacity])
            x = (tf.reshape(y, [-1, 1]) // div_num) % self.module
            x = tf.cast(x, 'int64')
            x = tf.reshape(x, [-1])
            size = int(np.prod(shape))
            x = tf.slice(x, [0], [size])
            inner_value = tf.reshape(x, shape)
        return SharedTensorBase(inner_value=inner_value, module=self.module, shape=self.shape)

    def identity(self):
        """
        The usage is same as tf.identity
        :return: A  copy of self
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.identity(self.inner_value)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def squeeze(self, axis=None):
        """
        The usage is same as tf.squeeze
        :param axis: An optional list of `ints`. Defaults to `[]`. If specified, only
                      squeezes the dimensions listed. The dimension index starts at 0. It is an
                      error to squeeze a dimension that is not 1. Must be in the range
                      `[-rank(input), rank(input))`. Must be specified if `input` is a
                      `RaggedTensor`.
        :return:  A `SharedTensorBase`. Has the same type as `self`.
                    Contains the same data as `self`, but has one or more dimensions of
                    size 1 removed.
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.squeeze(self.inner_value, axis=axis)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def rshift(self, index: tf.Tensor):
        """
        Cycle right shift self by index
        :param index: tf.Tensor where x.shape[:-1]==index.shape
        :return: y: SharedTensorBase of same module as self, where the vector y[i1,..., in-1,:] is the
                cycle right shift of x[i1,..., in-1,:] by index[i1, ...in-1] for any i1, ..., in-1
        """
        self.check_inner_value_is_not_None()
        if index.shape != self.shape[0:-1]:
            raise StfEqualException("index.shape", "self.shape[0:-1]", index.shape, self.shape[0:-1])
        index = index % self.shape[-1]
        shifted_reshape = cycle_rshift_tensor(tf.reshape(self.inner_value, shape=[-1, self.shape[-1]]),
                                              tf.reshape(index, shape=[-1]))
        return SharedTensorBase(inner_value=tf.reshape(shifted_reshape, shape=self.shape), module=self.module)

    def lshift(self, index: tf.Tensor):
        """
        Cycle right shift self by i
        :param index: tf.Tensor where x.shape[:-1]==index.shape
        :return: y: SharedTensorBase with same module as self, where the vector y[i1,..., in-1,:] is the
                cycle right shift of x[i1,..., in-1,:] by index[i1, ...in-1] for any i1, ..., in-1
        """
        self.check_inner_value_is_not_None()
        return self.rshift(-index)

    def reshape(self, shape):
        """

        :param shape: list, new shape
        :return:  reshaped SharedTensorBase
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.reshape(self.inner_value, shape)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def transpose(self):
        """

        :return: The transposed SharedTensorBase
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.transpose(self.inner_value)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def reduce_mean(self, axis, keepdims=False):
        """

        :param axis:  The dimensions to reduce. If `None` (the default), reduces all
      dimensions. Must be in the range `[-len(self.shape),
      len(self.shape))`.
        :param keepdims: If true, retains reduced dimensions with length 1.
        :return: The reduced tensor.
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.reduce_mean(self.inner_value, axis=axis, keepdims=keepdims)
        inner_value = tf.cast(inner_value, 'int64')
        if self.module is not None:
            inner_value %= self.module
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def concat(self, other, axis):
        """

        :param other: Another SharedTensorBase
        :param axis: Same as 'axis' in tf.concat
        :return: A `Tensor` resulting from concatenation of the self and other
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.concat(values=[self.inner_value, other.inner_value], axis=axis)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def stack(self, other, axis):
        """

        :param other:  Another SharedTensorBase
        :param axis: Same as 'axis' in tf.stack
        :return: A stacked SharedTensorBase with the same module of self
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.stack(values=[self.inner_value, other.inner_value], axis=axis)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def reduce_sum(self, axis, keepdims=False):
        """
        :param axis:  The dimensions to reduce. If `None` (the default), reduces all
      dimensions. Must be in the range `[-len(self.shape),
      len(self.shape))`.
        :param keepdims: If true, retains reduced dimensions with length 1.
        :return: The reduced tensor.
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.reduce_sum(self.inner_value, axis=axis, keepdims=keepdims)
        # inner_value = tf.cast(inner_value, 'int64')
        if self.module is not None:
            inner_value = inner_value % self.module
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def gather(self, indices, axis, batch_dims):
        """

        :param indices:  The index `Tensor`.  Must be one of the following types: `int32`,
      `int64`. Must be in range `[0, params.shape[axis])`.
        :param axis: A `Tensor`. Must be one of the following types: `int32`, `int64`. The
      `axis` in `params` to gather `indices` from. Must be greater than or equal
      to `batch_dims`.  Defaults to the first non-batch dimension. Supports
      negative indexes.
        :param batch_dims: An `integer`.  The number of batch dimensions.  Must be less
      than or equal to `rank(indices)`.
        :return: A `SharedTensorBase` with same module as self.
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.gather(params=self.inner_value, indices=indices, axis=axis, batch_dims=batch_dims)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def expand_dims(self, axis):
        """
        expand dimension of self, like tf.expand_dims.
        :param axis: Integer specifying the dimension index at which to expand the
      shape of `input`. Given an input of D dimensions, `axis` must be in range
      `[-(D+1), D]` (inclusive).
        :return: A `SharedTensorBase` with same module as self.
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.expand_dims(self.inner_value, axis=axis)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def cumulative_sum(self, axis=-1):
        """
        Compute the cumulative_sum of self along the axis
        :param axis: Integer specifying the dimension index at which to expand the
      shape of `input`. Given an input of D dimensions, `axis` must be in range
      `[-(D+1), D]` (inclusive).
        :return: SharedTensorBase of same shape and module as self
        """
        self.check_inner_value_is_not_None()
        perm = list(range(len(self.shape)))
        perm[0] = perm[axis]
        perm[axis] = 0
        x = tf.transpose(self.inner_value, perm=perm)
        y = tf.scan(lambda a, b: a + b, x, back_prop=False) % self.module
        y = tf.transpose(y, perm=perm)
        return SharedTensorBase(inner_value=y, module=self.module)

    def ones_like(self):
        """
        This operation returns a SharedTensorBase of the
        same module and shape as self with all elements set to 1.
        :return: A SharedTensorBase with all elements set to one.
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.ones_like(self.inner_value)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def complement(self):
        """
        For a SharedTensorBas with module 2, return the 1-self.
        :return: A SharedTensorBas with same shape as self and module 2.
        """
        self.check_inner_value_is_not_None()
        if self.module != 2:
            raise StfValueException("self.module", 2, self.module)
        inner_value = tf.math.logical_not(tf.cast(self.inner_value, 'bool'))
        inner_value = tf.cast(inner_value, 'int64')
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def zeros_like(self):
        """
        This operation returns a SharedTensorBase of the
        same module and shape as self with all elements set to 0.
        :return: A SharedTensorBase with all elements set to 0.

        """
        self.check_inner_value_is_not_None()
        inner_value = tf.zeros_like(self.inner_value)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def pad(self, paddings, mode="CONSTANT", constant_values=0):
        """

        :param paddings: A `Tensor` of type `int32`.
        :param mode: One of "CONSTANT", "REFLECT", or "SYMMETRIC" (case-insensitive)
        :param constant_values: In "CONSTANT" mode, the scalar pad value to use. Must has type 'int64'.
        :return:  A SharedTensorBase with same module as self.
        """
        self.check_inner_value_is_not_None()
        inner_value = tf.pad(self.inner_value, paddings=paddings, mode=mode, constant_values=constant_values)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def __add__(self, other):
        self.check_inner_value_is_not_None()
        self.check_module_equal(other)
        inner_value = (self.inner_value + other.inner_value)
        return SharedTensorBase(inner_value=inner_value, module=self.module)

    def __sub__(self, other):
        self.check_inner_value_is_not_None()
        self.check_module_equal(other)
        inner_value = (self.inner_value - other.inner_value)
        return SharedTensorBase(inner_value=inner_value,
                                module=self.module)

    def __neg__(self):
        self.check_inner_value_is_not_None()
        inner_value = - self.inner_value
        return SharedTensorBase(inner_value=inner_value,
                                module=self.module)

    def __mul__(self, other):
        if not isinstance(other, SharedTensorBase):
            raise StfTypeException("other", SharedTensorBase, type(other))
        self.check_inner_value_is_not_None()
        self.check_module_equal(other)
        inner_value = (self.inner_value * other.inner_value)
        return SharedTensorBase(inner_value=inner_value,
                                module=self.module)

    def __rmul__(self, other: Union[int, float, np.ndarray, tf.Tensor]):
        self.check_inner_value_is_not_None()
        inner_value = tf.cast(other, 'int64') * self.inner_value
        return SharedTensorBase(inner_value=inner_value,
                                module=self.module)

    def __truediv__(self, other: Union[int, float, np.ndarray, tf.Tensor]):
        self.check_inner_value_is_not_None()
        inner_value = tf.cast(self.inner_value / other, 'int64')
        return SharedTensorBase(inner_value=inner_value,
                                module=self.module)

    __floordiv__ = __truediv__

    def __matmul__(self, other):
        self.check_module_equal(other)
        self.check_inner_value_is_not_None()
        inner_value = tf.matmul(self.inner_value, other.inner_value)
        return SharedTensorBase(inner_value=inner_value,
                                module=self.module)

    def __lshift__(self, other: int):
        self.check_inner_value_is_not_None()
        if other > 62 or other < 0:
            raise StfCondException(cond="0<=other<=62", real="other={}".format(other))
        elif other == 0:
            return self
        else:  # other > 0
            inner_value = (1 << other) * self.inner_value
            return SharedTensorBase(inner_value=inner_value,
                                    module=self.module)

    def __rshift__(self, other: int):
        self.check_inner_value_is_not_None()
        if other < 0:
            raise StfCondException(cond="other>=0", real="other={}".format(other))
        elif other == 0:
            return self
        else:
            inner_value = self.inner_value // (1 << other)
            return SharedTensorBase(inner_value=inner_value,
                                    module=self.module)

    def __pow__(self, power: int):
        self.check_inner_value_is_not_None()
        if not isinstance(power, int):
            raise StfTypeException("power", 'int', type(power))
        elif power < 0:
            raise StfCondException(cond="power>=0", real="power={}".format(power))
        elif power == 0:
            return self.ones_like()
        elif power == 1:
            return self
        elif power == 2:
            inner_value = (self.inner_value ** 2) % self.module
            return SharedTensorBase(inner_value=inner_value, module=self.module)
        else:  # power > 2:
            power_half = power // 2
            y_power_half = self ** power_half
            if power % 2 == 0:
                return y_power_half ** 2
            else:
                return (y_power_half ** 2) * self


    def serialize(self, path=None):
        if path is None:
            raise StfNoneException("path")
        path = os.path.join(path, "serialize")
        if not os.path.exists(path):
            os.mkdir(path)
        path = os.path.join(path, self.name)

        y = tf.io.serialize_tensor(self.inner_value)
        w_op = tf.print(y, output_stream="file://"+path)
        StfConfig.pre_produce_list.append(w_op)
        self.serialize_len=tf.strings.length(tf.io.serialize_tensor(tf.zeros_like(self.inner_value)))

    def unserialize(self, path=None):
        shape = self.shape
        if path is None:
            raise StfNoneException("path")
        path = os.path.join(path, "serialize", self.name)
        # z = tf.compat.v1.data.TextLineDataset(
        #     filenames="file://"+path).make_one_shot_iterator().get_next()
        z = tf.compat.v1.data.FixedLengthRecordDataset(
            filenames="file://"+path, record_bytes=tf.cast(self.serialize_len, 'int64') + 1)\
            .repeat(StfConfig.offline_triple_multiplex).make_one_shot_iterator().get_next()
        z = tf.strings.substr(z, 0, self.serialize_len)
        self.inner_value = tf.reshape(tf.io.parse_tensor(z, 'int64'), shape)




def get_device(owner):
    if isinstance(owner, python.framework.device_spec.DeviceSpecV2):
        device = owner
    elif isinstance(owner, str):
        owner = str.split(owner, ":")  # "L:0"     "L:1"......      "R:0", "R:1"......
        if owner[0] == "L":
            if len(owner) == 1:
                device = StfConfig.workerL[0]
            else:
                device = StfConfig.workerL[int(owner[1])]
        elif owner[0] == "R":
            if len(owner) == 1:
                device = StfConfig.workerR[0]
            else:
                device = StfConfig.workerR[int(owner[1])]
        else:
            raise StfException(msg="owner error: owner={}".format(owner[0]))
    else:
        raise StfTypeException(obj_name="owner", expect_type="str or tf.python.framework.device_spec.DeviceSpecV2",
                               real_type=type(owner))
    return device


class PrivateTensorBase:
    """
    The Class of PrivateTensorBase
    inner_value: A Tensorflow Tensor or Variable of int64
    module: int，
    fixedpoint: int
    owner: String, "L" or "R"
    The represented number = inner_value * 2^{-fixedpoint}
    """

    def __init__(self, owner, fixedpoint: int = None, inner_value=None, module: int = None):
        if fixedpoint is None:
            fixedpoint = StfConfig.default_fixed_point
        if module is not None:
            if not isinstance(module, int):
                raise StfTypeException(obj_name="module", expect_type="int", real_type=type(module))
            else:
                if module <= 0:
                    raise StfCondException(cond="module<=0", real="module={}".format(module))

        self.module = module
        self.owner = get_device(owner)
        self.fixedpoint = fixedpoint
        if inner_value is not None:
            if self.module is not None:
                with tf.device(self.owner):
                    inner_value = inner_value % self.module
            if isinstance(inner_value, (tf.Tensor, tf.Variable)):
                if inner_value.device == self.owner.to_string():
                    self.inner_value = inner_value
                else:
                    with tf.device(self.owner):
                        self.inner_value = tf.identity(inner_value)
            else:
                with tf.device(self.owner):
                    self.inner_value = tf.constant(inner_value)
        self.name = ''.join(random.sample(string.ascii_letters + string.digits, 16))

    @property
    def shape(self):
        return self.inner_value.shape.as_list() if self.inner_value is not None else None

    def __repr__(self) -> str:
        return 'PrivateTensorBase(owner={}, fixedpoint={}, ' \
               'module={}, shape={})'.format(self.owner.to_string(), self.fixedpoint, self.module, self.shape)

    def check_inner_value_is_not_None(self):
        if self.inner_value is None:
            raise StfNoneException(obj_name="self.inner_value")

    def check_module_equal(self, other):
        if self.module != other.module:
            raise StfEqualException("self.module", "other.module", self.module, other.module)

    def check_owner_equal(self, other):
        if self.owner != other.owner:
            raise StfEqualException("self.owner", "other.owner", self.owner, other.owner)


    def slice(self, begin, size):
        self.check_inner_value_is_not_None()
        with tf.device(self.owner):
            inner_value = tf.slice(self.inner_value, begin, size)
        return PrivateTensorBase(self.owner, fixedpoint=self.fixedpoint, inner_value=inner_value, module=self.module)

    def split(self, size_splits, axis=0, num=None):
        self.check_inner_value_is_not_None()
        with tf.device(self.owner):
            inner_values = tf.split(self.inner_value, size_splits, axis, num)
            pvs = []
            for inner_value in inner_values:
                pvs += [PrivateTensorBase(self.owner, fixedpoint=self.fixedpoint,
                                          inner_value=inner_value, module=self.module)]
        return tuple(pvs)

    def squeeze(self, axis):
        inner_value = tf.squeeze(self.inner_value, axis=axis)
        return PrivateTensorBase(self.owner, fixedpoint=self.fixedpoint,
                                          inner_value=inner_value, module=self.module)

    def load_from_tf_tensor(self, x: Union[tf.Tensor, np.ndarray]):
        """
        Load from tensorflow tensor
        :param x: tf.Tensor
        :return:
        """
        with tf.device(self.owner):
            self.inner_value = tf.cast(tf.multiply(x, (1 << self.fixedpoint)), 'int64')
            if self.module is not None:
                self.inner_value %= self.module
            if self.shape is not None:
                if self.shape != self.inner_value.shape:
                    raise StfEqualException("self.shape", "self.inner_value.shape",
                                            self.shape, self.inner_value.shape)

    def identity(self):
        """
        See tf.identity.
        :return:
        """
        self.check_inner_value_is_not_None()
        with tf.device(self.owner):
            inner_value = tf.identity(self.inner_value)
        return PrivateTensorBase(self.owner, fixedpoint=self.fixedpoint, inner_value=inner_value, module=self.module)

    def load_from_file(self, path: str, record_defaults, batch_size, field_delim=",", skip_row_num=1, skip_col_num=0,
                       repeat=1, clip_value=None, scale=1.0, map_fn=None, output_col_num=None, buffer_size=0):
        """
        Load data from file
        :param path:  absolute path of file in the disk of self.owner
        :param record_defaults: for example [['a'], ['a'], [1.0], [1.0], [1.0]]
        :param batch_size:
        :param field_delim:    field delim between columns
        :param skip_row_num:   skip row number in head of the file
        :param skip_col_num:   skip column number in the file
        :param repeat:         repeat how many times of the file
        :param clip_value:     the features are clip by this value such that |x|<=clip_value
        :param scale:          multiply scale for the  columns
        :param map_fn:         A map function for the columns, for example: lambda x: x[3]*x[4]
        :param output_col_num:   output column number
        :param buffer_size:       buffer size
        :return:
        """
        def clip(r):
            if clip_value is None:
                return r * scale if scale != 1.0 else r
            else:
                return tf.clip_by_value(r * scale, -clip_value, clip_value)

        if output_col_num is None:
            output_col_num = len(record_defaults) - skip_col_num

        with tf.device(self.owner):
            data = tf.compat.v1.data.TextLineDataset(path, buffer_size=buffer_size).skip(skip_row_num)
            data_iter = data.repeat(repeat).batch(batch_size).make_one_shot_iterator()
            data = data_iter.get_next()
            data = tf.reshape(data, [batch_size])
            data = tf.strings.split(data, sep=field_delim).to_tensor(default_value="0.0")
            data = data[:, skip_col_num:]
            data = tf.reshape(data, [batch_size, output_col_num])
            data = tf.strings.to_number(data, out_type='float64')
            data = clip(data)
            if map_fn is not None:
                data = data.map(map_func=map_fn)
        self.load_from_tf_tensor(data)

    def __getitem__(self, item):
        with tf.device(self.owner):
            inner_value = self.inner_value.__getitem__(item)
        return PrivateTensorBase(self.owner, fixedpoint=self.fixedpoint, inner_value=inner_value,
                                 module=self.module)

    def load_from_file_withid(self, path: str, record_defaults, batch_size, field_delim=",",
                              skip_row_num=1, id_col_num=0, repeat=1, clip_value=None, use_auto_id_col=False):
        """
        Load data from file, and return the id columns.
        :param path:  absolute path of file in the disk of self.owner
        :param record_defaults: for example [['a'], ['a'], [1.0], [1.0], [1.0]]
        :param batch_size:
        :param field_delim:    field delim between columns
        :param skip_row_num:   skip row number in head of the file
        :param repeat:         repeat how many times of the file
        :param clip_value:     the features are clip by this value such that |x|<=clip_value
        :param id_col_num:          the number of front columns that are id (not feature or label)
        :param use_auto_id_col:  if true, add a auto increase id column [0, 1, 2, ......]
        :return: tf.Tensor of id columns
        """

        def clip(r):
            if clip_value is None:
                return r
            else:
                return tf.clip_by_value(r, -clip_value, clip_value)

        with tf.device(self.owner):
            output_col_num = len(record_defaults) - id_col_num
            data = tf.compat.v1.data.TextLineDataset(path).skip(skip_row_num)
            data_iter = data.repeat(repeat).batch(batch_size).make_one_shot_iterator()
            data = data_iter.get_next()
            data = tf.reshape(data, [batch_size])
            data = tf.strings.split(data, sep=field_delim).to_tensor(default_value="0.0")

            if use_auto_id_col:
                max_value = StfConfig.upper_bd_int64
                idx = tf.compat.v1.data.Dataset.range(max_value).batch(batch_size)
                idx_iter = idx.make_one_shot_iterator()
                idx_batch = tf.reshape(idx_iter.get_next(), [batch_size, id_col_num])
            else:
                idx_batch = data[:, 0:id_col_num]
            data = data[:, id_col_num:]
            data = tf.reshape(data, [batch_size, output_col_num])
            data = tf.strings.to_number(data, out_type='float64')
            data = clip(data)

        self.load_from_tf_tensor(data)
        return idx_batch

    def load_first_line_from_file(self, path: str, col_num, sep=",", dtype='float32'):
        """
        Load the first line of the file
        :param path:  absolute path of file in the disk of self.owner
        :param col_num: columns number of the file
        :param sep:    separation character between columns
        :param dtype:  tf.dtype
        :return:
        """
        with tf.device(self.owner):
            aline = tf.io.read_file(filename=path)
            aline = tf.strings.split(aline, sep="\n")
            aline = aline[0]
            aline = tf.strings.split(aline, sep=sep)
            aline = tf.strings.to_number(aline, dtype)
            aline = tf.reshape(aline, [col_num])
            self.load_from_tf_tensor(aline)

    def to_tf_tensor(self, owner=None, dtype='float64') -> tf.Tensor:
        """
        Transform self to a tf.Tensor
        :param owner:  which machine the tf.Tensor lie
        :param dtype:  tf.Dtype  'float64' or 'int64'
        :return:      tf.Tensor
        """
        self.check_inner_value_is_not_None()
        if owner is None:
            owner = self.owner
        else:
            owner = get_device(owner)
        if owner != self.owner:
            # raise StfEqualException("owner", "self.owner", owner, self.owner)
            StfEqualWarning("owner", "self.owner", owner, self.owner)
        with tf.device(owner):
            if dtype == 'float64':
                return tf.cast(tf.cast(self.inner_value, 'float64') / (2 ** self.fixedpoint), 'float64')
            elif dtype == 'int64':
                return self.inner_value // (2 ** self.fixedpoint)
            else:
                raise StfValueException("dtype", "float64 or int64", dtype)

    def to_tf_str(self, owner=None, precision=StfConfig.to_str_precision, id_col=None):
        """
        Transform self to a tf.Tensor of dtype string.
        :param owner:  which machine the tf.Tensor lie
        :param precision:  An optional `int`. Defaults to `StfConfig.to_str_precision`.
      The post-decimal precision to use for floating point numbers.
      Only used if precision > -1.
        :param id_col: if not None, add a id column to the data.
        :return: tr.Tensor of dtype string
        """
        self.check_inner_value_is_not_None()
        if owner is None:
            owner = self.owner
        x = self.to_tf_tensor(owner=owner)
        with tf.device(get_device(owner)):
            y = tf.strings.as_string(x, precision=precision)
            if id_col is not None:
                y = tf.concat([id_col, y], axis=1)
            y = tf.compat.v1.reduce_join(y, separator=",", axis=-1)
            return y

    def to_file(self, path: str, separator=",", precision=StfConfig.to_str_precision, dim: int = 1, owner=None,
                dtype='float64', id_col=None):
        """
        Generate a tf.operator, when run it, self is writen to a file in the machine of owner.
        :param path: The file to be written in the machine of self.owner.
        :param separator:  separation character between columns
        :param precision: An optional `int`. Defaults to `StfConfig.to_str_precision`.
      The post-decimal precision to use for floating point numbers.
      Only used if precision > -1.
        :param dim:   the dimension of self
        :param owner:  'L' or 'R'
        :param dtype:  tf.dtype  'int64' or 'float64'
        :param id_col: if not None, add a id column to the data.
        :return:
        """
        self.check_inner_value_is_not_None()
        if dim > 2 or dim < 1:
            raise StfValueException("dim", "1 or 2", dim)
        if owner is None:
            owner = self.owner
        else:
            owner = get_device(owner)

        with tf.device(owner):
            tf_tensor = self.to_tf_tensor(dtype=dtype)
            if dtype == 'float64':
                tf_tensor = tf.strings.as_string(tf_tensor, precision=precision)
            elif dtype == 'int64':
                tf_tensor = tf.strings.as_string(tf_tensor)
            else:
                raise StfValueException("dtype", 'float64 or int64', dtype)

            if dim == 2 and id_col is not None:
                tf_tensor = tf.concat(values=[id_col, tf_tensor], axis=1)

            weights = tf.compat.v1.reduce_join(tf_tensor, separator=separator, axis=-1)
            if dim == 2:
                weights = tf.compat.v1.reduce_join(weights, separator="\n", axis=-1)
            write_op = tf.io.write_file(filename=path, contents=weights)
            return write_op

    def to_SharedPairBase(self, other_owner):
        """

        :param other_owner:
        :param op_map:
        :return:

        """
        xL = SharedTensorBase(inner_value=self.inner_value, module=self.module)
        xR = -xL.ones_like()  # use one but not zero, for div
        xL += xL.ones_like()
        x = SharedPairBase(ownerL=self.owner, ownerR=other_owner, xL=xL, xR=xR, fixedpoint=self.fixedpoint)
        return x


    def dup_with_precision(self, new_fixedpoint: int):
        """
        Duplicate self to a new fixedpoint.
        :param new_fixedpoint:
        :return:  PrivateTensorBase
        """
        self.check_inner_value_is_not_None()
        if new_fixedpoint == self.fixedpoint:
            return self
        with tf.device(self.owner):
            if new_fixedpoint > self.fixedpoint:
                inner_value = self.inner_value * (1 << (new_fixedpoint - self.fixedpoint))  # left shift
            else:
                inner_value = self.inner_value // (1 << (self.fixedpoint - new_fixedpoint))
            return PrivateTensorBase(self.owner, fixedpoint=new_fixedpoint, inner_value=inner_value,
                                     module=self.module)

    def gather(self, indices, axis, batch_dims):
        """ See tf.gather."""
        self.check_inner_value_is_not_None()
        with tf.device(self.owner):
            inner_value = tf.gather(params=self.inner_value, indices=indices, axis=axis, batch_dims=batch_dims)
        return PrivateTensorBase(self.owner, fixedpoint=self.fixedpoint, inner_value=inner_value,
                                 module=self.module)

    def stack(self, other, axis):
        """See tf.stack."""
        self.check_inner_value_is_not_None()
        self.check_module_equal(other)
        self.check_owner_equal(other)

        fixed_point = min(self.fixedpoint, other.fixedpoint)
        x = self.dup_with_precision(fixed_point)
        y = other.dup_with_precision(fixed_point)
        with tf.device(self.owner):
            inner_value = tf.stack([x.inner_value, y.inner_value], axis=axis)
        return PrivateTensorBase(self.owner, fixedpoint=fixed_point, inner_value=inner_value, module=self.module)

    def concat(self, other, axis):
        """See tf.concat"""
        self.check_inner_value_is_not_None()
        self.check_module_equal(other)
        self.check_owner_equal(other)

        fixed_point = min(self.fixedpoint, other.fixedpoint)
        x = self.dup_with_precision(fixed_point)
        y = other.dup_with_precision(fixed_point)
        with tf.device(self.owner):
            inner_value = tf.concat([x.inner_value, y.inner_value], axis=axis)
        return PrivateTensorBase(self.owner, fixedpoint=fixed_point, inner_value=inner_value, module=self.module)

    def reshape(self, shape):
        """See tf.reshape."""
        self.check_inner_value_is_not_None()
        with tf.device(self.owner):
            inner_value = tf.reshape(self.inner_value, shape)
            return PrivateTensorBase(owner=self.owner, fixedpoint=self.fixedpoint, inner_value=inner_value,
                                     module=self.module)

    def random_uniform_adjoint(self, seed=None):
        """
        generate random uniform adjoint of self.
        :param seed: A seed get from random.get_seed()
        :return:  A uniform random PrivateTensor with same owner, module and fixedpoint as self.
        the value is differential in every time sess.run it
        """
        with tf.device(self.owner):
            x = self.to_SharedTensor_like()
            y = x.random_uniform_adjoint(seed=seed)
            adjoint = PrivateTensorBase(owner=self.owner, fixedpoint=self.fixedpoint, inner_value=y.inner_value,
                                        module=self.module)
            return adjoint

    def zeros_like(self):
        """See tf.zeros_like."""
        with tf.device(self.owner):
            inner_value = tf.zeros_like(self.inner_value)
        return PrivateTensorBase(owner=self.owner, fixedpoint=self.fixedpoint, inner_value=inner_value,
                                 module=self.module)

    def ones_like(self, fixedpoint=None):
        """See tf.ones_like."""
        if fixedpoint is None:
            fixedpoint = self.fixedpoint
        with tf.device(self.owner):
            inner_value = (1 << fixedpoint) * tf.ones_like(self.inner_value)
        return PrivateTensorBase(owner=self.owner, fixedpoint=fixedpoint, inner_value=inner_value,
                                 module=self.module)

    def to_SharedTensor(self):
        """
        :return a SharedTensorBase with same inner_value and module with self.
        """
        return SharedTensorBase(inner_value=self.inner_value, module=self.module)

    def to_SharedTensor_like(self):
        """
        :return a SharedTensorBase with inner_value=None, and with same module and shape with self.
        """
        return SharedTensorBase(module=self.module, shape=self.shape)

    def serialize(self, path=None):
        if path is None:
            path = StfConfig.stf_home_workerL if self.owner == StfConfig.workerL[0] else StfConfig.stf_home_workerR
        # print("line 1015 base.py, path=", path)
        path = os.path.join(path, "serialize")
        if not os.path.exists(path):
            os.mkdir(path)
        path = os.path.join(path, self.name)
        # print("line 1015 base.py, path=", path)
        with tf.device(self.owner):
            y = tf.io.serialize_tensor(self.inner_value)
            self.serialize_len = tf.strings.length(tf.io.serialize_tensor(tf.zeros_like(self.inner_value)))
            w_op = tf.print(y, output_stream="file://"+path)
        StfConfig.pre_produce_list.append(w_op)
        # print("StfConfig.pre_produce_list=", StfConfig.pre_produce_list)



    def unserialize(self, path=None):
        shape =self.shape
        if path is None:
            path = StfConfig.stf_home_workerL if self.owner == StfConfig.workerL[0] else StfConfig.stf_home_workerR
        path = os.path.join(path, "serialize", self.name)
        with tf.device(self.owner):
            # z = tf.compat.v1.data.TextLineDataset(
            #     filenames="file://"+path).make_one_shot_iterator().get_next()
            z = tf.compat.v1.data.FixedLengthRecordDataset(
                filenames="file://" + path,
                record_bytes=tf.cast(self.serialize_len, 'int64') + 1).repeat(StfConfig.offline_triple_multiplex).make_one_shot_iterator().get_next()
            z = tf.strings.substr(z, 0, self.serialize_len)
            self.inner_value = tf.reshape(tf.io.parse_tensor(z, 'int64'), shape)
        self.inner_value = tf.reshape(tf.io.parse_tensor(z, 'int64'), shape)


    def __add__(self, other):
        self.check_inner_value_is_not_None()
        self.check_module_equal(other)
        self.check_owner_equal(other)
        with tf.device(self.owner):
            fixed_point = min(self.fixedpoint, other.fixedpoint)
            altered_self = self.dup_with_precision(new_fixedpoint=fixed_point)
            altered_other = other.dup_with_precision(new_fixedpoint=fixed_point)
            inner_value = (altered_self.inner_value + altered_other.inner_value)
            if self.module is not None:
                inner_value %= self.module
            return PrivateTensorBase(owner=self.owner, fixedpoint=fixed_point, inner_value=inner_value,
                                     module=self.module)

    def __neg__(self):
        self.check_inner_value_is_not_None()
        with tf.device(self.owner):
            inner_value = - self.inner_value
            if self.module is not None:
                inner_value = inner_value % self.module
            return PrivateTensorBase(owner=self.owner, fixedpoint=self.fixedpoint, inner_value=inner_value,
                                     module=self.module)

    def __sub__(self, other):
        return self.__add__(other.__neg__())

    def __mul__(self, other, fixed_point=None):

        if self.owner != other.owner:
            raise Exception("owner must be same")
        if self.module != other.module:
            raise Exception("module must be same")
        with tf.device(self.owner):
            inner_value = (self.inner_value * other.inner_value)
            if self.module is not None:
                inner_value = inner_value % self.module
            w = PrivateTensorBase(owner=self.owner, fixedpoint=self.fixedpoint + other.fixedpoint,
                                  inner_value=inner_value,
                                  module=self.module)
        if not fixed_point:
            fixed_point = max(self.fixedpoint, other.fixedpoint)
            # fixed_point = self.fixedpoint + other.fixedpoint
        return w.dup_with_precision(fixed_point)

    def __matmul__(self, other, fixed_point=None):
        if self.owner != other.owner:
            raise Exception("owner must be same")
        if self.module != other.module:
            raise Exception("module must be same")
        with tf.device(self.owner):
            inner_value = tf.matmul(self.inner_value, other.inner_value)
            if self.module is not None:
                inner_value = inner_value % self.module
            w = PrivateTensorBase(owner=self.owner, fixedpoint=self.fixedpoint + other.fixedpoint,
                                  inner_value=inner_value,
                                  module=self.module)
        if not fixed_point:
            fixed_point = max(self.fixedpoint, other.fixedpoint)
            # fixed_point = self.fixedpoint + other.fixedpoint
        return w.dup_with_precision(fixed_point)

    def __lt__(self, other):
        if self.owner != other.owner:
            raise Exception("owner must be same")
        if self.module != other.module:
            raise Exception("module must be same")
        with tf.device(self.owner):
            inner_value = tf.cast(self.to_tf_tensor() < other.to_tf_tensor(), 'int64')
        return PrivateTensorBase(owner=self.owner, fixedpoint=0, inner_value=inner_value, module=2)

    def __le__(self, other):
        if self.owner != other.owner:
            raise Exception("owner must be same")
        if self.module != other.module:
            raise Exception("module must be same")
        with tf.device(self.owner):
            inner_value = tf.cast(self.to_tf_tensor() <= other.to_tf_tensor(), 'int64')
        return PrivateTensorBase(owner=self.owner, fixedpoint=0, inner_value=inner_value, module=2)

    def __gt__(self, other):
        if self.owner != other.owner:
            raise Exception("owner must be same")
        if self.module != other.module:
            raise Exception("module must be same")
        with tf.device(self.owner):
            inner_value = tf.cast(self.to_tf_tensor() > other.to_tf_tensor(), 'int64')
        return PrivateTensorBase(owner=self.owner, fixedpoint=0, inner_value=inner_value, module=2)

    def __ge__(self, other):
        if self.owner != other.owner:
            raise Exception("owner must be same")
        if self.module != other.module:
            raise Exception("module must be same")
        with tf.device(self.owner):
            inner_value = tf.cast(self.to_tf_tensor() >= other.to_tf_tensor(), 'int64')
        return PrivateTensorBase(owner=self.owner, fixedpoint=0, inner_value=inner_value, module=2)


class SharedPairBase:
    """
    A SharedPairBase x is represented by x= (xL+xR mod n)*2^{-fixedpoint},
    where xL is in ownerL and xR is in ownerR, n=xL.module=xR.module.
    :param ownerL: str (for example, "L") or object of python.framework.device_spec.DeviceSpecV2
    :param ownerR: str (for example, "R") or object of python.framework.device_spec.DeviceSpecV2
    :param xL: a SharedTensorBase.
    :param xR: a SharedTensorBase.
    :param fixedpoint: int. default is StfConfig.default_fixed_point
    :param shape:  list of int
    Shape or (xL,xR) must not be None.
    """
    def __init__(self, ownerL, ownerR, xL: SharedTensorBase = None, xR: SharedTensorBase = None,
                 fixedpoint: int = StfConfig.default_fixed_point,
                 shape=None):
        ownerL = get_device(ownerL)
        ownerR = get_device(ownerR)

        if ownerL == ownerR:
            raise StfException(
                "The ownerL and ownerR should be different. But ownerL={}, ownerR={}".format(ownerL, ownerR))
        if shape is not None and xL is not None:
            if shape != xL.shape:
                raise StfEqualException("shape", "xL.shape", shape, xL.shape)
        self.ownerL = ownerL
        self.ownerR = ownerR
        self.fixedpoint = fixedpoint
        if xL is not None:
            if xL.device == self.ownerL.to_string():
                self.xL = xL
            else:
                with tf.device(self.ownerL):
                    self.xL = xL.identity()
        else:
            self.xL = None
        if xR is not None:
            if xR.device == self.ownerR.to_string():
                self.xR = xR
            else:
                with tf.device(self.ownerR):
                    self.xR = xR.identity()
        else:
            self.xR = None
        if xL is not None and xR is not None:
            if xL.module != xR.module:
                raise StfEqualException("xL.module", "xR.module", xL.module, xR.module)
        if shape is not None:
            self.shape = shape
        elif self.xL is not None:
            self.shape = self.xL.shape
        else:
            raise StfNoneException("shape or xL")

    def mirror(self):
        """
        Get the mirrored SharedPairBase of self.
        :return:
        """
        ownerR = self.ownerL
        ownerL = self.ownerR
        xR = self.xL
        xL = self.xR
        return SharedPairBase(ownerL, ownerR, xL, xR, self.fixedpoint)

    def __repr__(self) -> str:
        return 'SharedPairBase(ownerL={},ownerR={}, fixedpoint={}, module={}, shape={})'.format(self.ownerL.to_string(),
                                                                                                self.ownerR.to_string(),
                                                                                                self.fixedpoint,
                                                                                                self.xL.module,
                                                                                                self.shape)

    def __getitem__(self, index):
        xL = self.xL[index]
        xR = self.xR[index]
        z = SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=self.fixedpoint)
        return z

    def split(self, size_splits, axis: int = 0, num=None):
        """
        See tf.split.
        :param size_splits:
        :param axis:
        :param num:
        :return:
        """
        with tf.device(self.ownerL):
            xLs = self.xL.split(size_splits=size_splits, axis=axis, num=num)
        with tf.device(self.ownerR):
            xRs = self.xR.split(size_splits=size_splits, axis=axis, num=num)
        sps = []
        for xL_xR in zip(xLs, xRs):
            sps += [SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR,
                                   xL=xL_xR[0], xR=xL_xR[1], fixedpoint=self.fixedpoint)]
        return tuple(sps)

    def to_SharedTensor_like(self):
        """
        :return:  A SharedTensorBase with inner_value=None and same module and shape with self.
        """
        return SharedTensorBase(module=self.xR.module, shape=self.shape)

    def identity(self, ownerL=None, ownerR=None):
        """
        See tf.identity.
        :param ownerL:
        :param ownerR:
        :return:
        """
        if ownerL is None:
            ownerL = self.ownerL
        if ownerR is None:
            ownerR = self.ownerR
        with tf.device(ownerL):
            xL = self.xL.identity()
        with tf.device(ownerR):
            xR = self.xR.identity()
        return SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=self.fixedpoint)

    def to_private(self, owner) -> PrivateTensorBase:
        """
        transform to a PrivateTensorBase.
        :param owner:
        :return:
        """
        with tf.device(get_device(owner)):
            x = self.xL.inner_value + self.xR.inner_value
            return PrivateTensorBase(owner=owner, fixedpoint=self.fixedpoint, inner_value=x, module=self.xL.module)

    def to_tf_tensor(self, owner) -> tf.Tensor:
        """
        transform self to a tensorflow Tensor.
        :param owner:
        :return:
        """
        with tf.device(get_device(owner)):
            x = self.to_private(owner)
            return x.to_tf_tensor()

    def ones_like(self):
        """
        See tf.ones_like
        :return:
        """
        with tf.device(self.ownerL):
            xL = self.xL.zeros_like()
        with tf.device(self.ownerR):
            xR = self.xR.ones_like()  # << self.fixedpoint
        return SharedPairBase(self.ownerL, self.ownerR, xL, xR, fixedpoint=0)

    def complement(self):
        """
        compute the two's complement of self.
        :return:
        """
        with tf.device(self.ownerL):
            xL = self.xL.complement()
        return SharedPairBase(self.ownerL, self.ownerR, xL, self.xR, fixedpoint=self.fixedpoint)

    def zeros_like(self):
        """
        See tf.zeros_like.
        :return:
        """
        with tf.device(self.ownerL):
            xL = self.xL.zeros_like()
        with tf.device(self.ownerR):
            xR = self.xR.zeros_like()
        return SharedPairBase(self.ownerL, self.ownerR, xL, xR, fixedpoint=self.fixedpoint)

    def cumulative_sum(self, axis=-1):
        """
        Compute the cumulative_sum of self alone the given axis.
        :param axis:
        :return:
        """
        with tf.device(self.ownerL):
            xL = self.xL.cumulative_sum(axis)
        with tf.device(self.ownerR):
            xR = self.xR.cumulative_sum(axis)
        return SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=self.fixedpoint)

    def load_from_tf_tensor(self, x):
        """
        Load data from a tensorflow Tensor.
        :param x:
        :return:
        """
        with tf.device(self.ownerL):
            px = PrivateTensorBase(owner=self.ownerL, fixedpoint=self.fixedpoint)
            px.load_from_tf_tensor(x)
            u = px.to_SharedTensor()
            self.xL = -u.ones_like()
        with tf.device(self.ownerR):
            self.xR = u + u.ones_like()
        self.fixedpoint = px.fixedpoint

    def _align_owner(self, other):
        """
        Align the owner of other with self.
        :param other: A SharedPairBase.
        :return: A SharedPairBase with same value as `other` satisfying ownerL=self.ownerL
        and ownerR=self.ownerR.
        """
        if self.ownerL == other.ownerL and self.ownerR == other.ownerR:
            return other
        elif self.ownerL == other.ownerR and self.ownerR == other.ownerL:
            return other.mirror()
        else:
            raise Exception("ownerL must be same, ownerR must be same.")

    def concat(self, other, axis):
        """
        See tf.concat.
        :param other:
        :param axis:
        :return:
        """
        other = self._align_owner(other)
        fixed_point = min(self.fixedpoint, other.fixedpoint)
        alter_self = self.dup_with_precision(fixed_point)
        other = other.dup_with_precision(fixed_point)
        with tf.device(self.ownerL):
            xL = alter_self.xL.concat(other.xL, axis=axis)
        with tf.device(self.ownerR):
            xR = alter_self.xR.concat(other.xR, axis=axis)
        return SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=fixed_point)

    def stack(self, other, axis):
        """
        See tf.stack.
        :param other:
        :param axis:
        :return:
        """
        other = self._align_owner(other)
        fixed_point = min(self.fixedpoint, other.fixedpoint)
        alter_self = self.dup_with_precision(fixed_point)
        other = other.dup_with_precision(fixed_point)
        with tf.device(self.ownerL):
            xL = alter_self.xL.stack(other.xL, axis=axis)
        with tf.device(self.ownerR):
            xR = alter_self.xR.stack(other.xR, axis=axis)
        return SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=fixed_point)

    def dup_with_precision(self, new_fixedpoint: int):
        """
        Duplicate self to a new SharedPairBase with new fixedpoint.
        :param new_fixedpoint:
        :return:
        """
        if new_fixedpoint > self.fixedpoint:
            with tf.device(self.ownerL):
                xL = self.xL << (new_fixedpoint - self.fixedpoint)  # left shift
            with tf.device(self.ownerR):
                xR = self.xR << (new_fixedpoint - self.fixedpoint)  # left shift
            return SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=new_fixedpoint)
        elif new_fixedpoint < self.fixedpoint:
            with tf.device(self.ownerL):
                xL = self.xL >> (self.fixedpoint - new_fixedpoint)  # right shift
            with tf.device(self.ownerR):
                xR = -((-self.xR) >> (self.fixedpoint - new_fixedpoint))  # right shift
            return SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=new_fixedpoint)
        else:
            return self

    def reshape(self, shape):
        """
        See tf.reshape.
        :param shape:
        :return:
        """
        with tf.device(self.ownerL):
            xL = self.xL.reshape(shape)
        with tf.device(self.ownerR):
            xR = self.xR.reshape(shape)
        return SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=self.fixedpoint)

    def squeeze(self, axis=None):
        """
        See tf.squeeze.
        :param axis:
        :return:
        """
        with tf.device(self.ownerL):
            xL = self.xL.squeeze(axis=axis)
        with tf.device(self.ownerR):
            xR = self.xR.squeeze(axis=axis)
        return SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=self.fixedpoint)

    def to_tf_str(self, owner, precision=StfConfig.to_str_precision, id_col=None):
        """
        Transform self to tf.string.
        :param owner: The owner of tf
        :param precision:
        :param id_col:
        :return:
        """
        return self.to_private(owner=owner).to_tf_str(precision=precision, id_col=id_col)

    def serialize(self):
        with tf.device(StfConfig.workerL[0]):
            self.xL.serialize(StfConfig.stf_home_workerL)
        with tf.device(StfConfig.workerR[0]):
            self.xR.serialize(StfConfig.stf_home_workerR)

    def unserialize(self):
        with tf.device(StfConfig.workerL[0]):
            self.xL.unserialize(StfConfig.stf_home_workerL)
        with tf.device(StfConfig.workerR[0]):
            self.xR.unserialize(StfConfig.stf_home_workerR)

    def random_uniform_adjoint(self, seed=None):
        with tf.device(self.ownerL):
            xL = self.xL.random_uniform_adjoint(seed)
        with tf.device(self.ownerR):
            xR = self.xR.random_uniform_adjoint(seed)
        z = SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR,
                           fixedpoint=self.fixedpoint)
        return z


    def __add__(self, other):
        other = self._align_owner(other)
        fixed_point = min(self.fixedpoint, other.fixedpoint)
        altered_self = self.dup_with_precision(new_fixedpoint=fixed_point)
        altered_other = other.dup_with_precision(new_fixedpoint=fixed_point)
        with tf.device(self.ownerL):
            xL = altered_self.xL + altered_other.xL
        with tf.device(self.ownerR):
            xR = altered_self.xR + altered_other.xR

        return SharedPairBase(self.ownerL, self.ownerR, xL, xR, fixed_point)

    def __neg__(self):
        return SharedPairBase(self.ownerL, self.ownerR, -self.xL, -self.xR, self.fixedpoint)

    def __sub__(self, other):
        return self + (-other)

    def __rmul__(self, other):
        with tf.device(self.ownerL):
            xL = other * self.xL
        with tf.device(self.ownerR):
            xR = other * self.xR
        return SharedPairBase(self.ownerL, self.ownerR, xL, xR, self.fixedpoint)



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : logical
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-05-14 11:42
   Description : description what the main function of this file
"""
from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedPairBase
from stensorflow.basic.operator.arithmetic import mul, add
from typing import Union
from stensorflow.global_var import StfConfig


def check_type_module(x, y):
    if isinstance(x, PrivateTensorBase) and isinstance(y, PrivateTensorBase):
        if x.module != 2 or y.module != 2:
            raise Exception("both x and y must have module 2")
    elif isinstance(x, PrivateTensorBase) and isinstance(y, SharedPairBase):
        if x.module != 2 or y.xL.module != 2:
            raise Exception("both x and y must have module 2")
    elif isinstance(x, SharedPairBase) and isinstance(y, PrivateTensorBase):
        if x.xL.module != 2 or y.module != 2:
            raise Exception("both x and y must have module 2")
    elif isinstance(x, SharedPairBase) and isinstance(y, SharedPairBase):
        if x.xL.module != 2 or y.xL.module != 2:
            raise Exception("both x and y must have module 2")
    else:
        raise Exception("x, y must be PrivateTensorBase or SharedPairBase")


def and_op(x, y, prf_flag=None):
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    check_type_module(x, y)
    return mul(x, y, prf_flag=prf_flag)


def xor_op(x, y):
    check_type_module(x, y)
    return add(x, y)


def or_op(x, y, prf_flag=None):
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    check_type_module(x, y)
    return add(add(x, y), mul(x, y, prf_flag=prf_flag))


def not_op(x: Union[PrivateTensorBase, SharedPairBase]):
    if isinstance(x, PrivateTensorBase):
        if x.module != 2:
            raise Exception("must have x.module == 2")
    elif isinstance(x, SharedPairBase):
        if x.xL.module != 2:
            raise Exception("must have x.module == 2")
    else:
        raise Exception("x must be a PrivateTensorBase or SharedPairBase")
    return x.ones_like() - x


def nxor_op(x, y):
    z = xor_op(x, y)
    return not_op(z)



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : float_compare
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-09-24 19:54
   Description : description what the main function of this file
   export PYTHONPATH=$PYTHONPATH:/Users/qizhi.zqz/projects/morse-stf/morse-stf
"""
import numpy as np
import pandas as pd
import argparse
import tensorflow as tf
from stensorflow.basic.basic_class.private import PrivateTensor
from stensorflow.basic.basic_class.pair import SharedPair
from stensorflow.basic.basic_class.base import SharedTensorBase, SharedPairBase, get_device
# from stensorflow.basic.protocol.compare import geq_SharedTensorBase_SharedTensorBase, \
#     greater_SharedTensorBase_SharedTensorBase, less_SharedTensorBase_SharedTensorBase, \
#     leq_SharedTensorBase_SharedTensorBase
from stensorflow.basic.protocol.msb import to_bool, bool_expansion, PrivateTensorBitwise, SharedTensorBitwise
from typing import Union
import time as time

def geq(x: tf.Tensor, y: tf.Tensor, x_owner, y_owner) -> SharedPair:
    """
     x>=y in Z/2^nZ iff  MSB(x+(2^n+2^n-y) in Z/2^{n+1}Z)==0  iff MSB(x+(2^n-1-y)+1 in Z/2^{n+1}Z)==1
    :param x:
    :param y:
    :param x_owner:
    :param y_owner:
    :return:
    """


    if x.dtype==tf.int16 or x.dtype==tf.int64 or x.dtype==tf.int32 or x.dtype==tf.float64 or x.dtype==tf.float32 or x.dtype==tf.float16:
        with tf.device(x_owner):
            list_x = bool_expansion(x)
            list_x = list(map(lambda xi: PrivateTensorBitwise(owner=x_owner, inner_value=SharedTensorBitwise(inner_value=xi)), list_x))

        with tf.device(y_owner):
            #list_y = bool_expansion(-tf.ones_like(y)-y)
            list_y = bool_expansion(y)
            list_y = list(map(lambda xi: PrivateTensorBitwise(owner=y_owner, inner_value=SharedTensorBitwise(inner_value=-tf.ones_like(xi)-xi)), list_y))

        carry = list_x[0] + list_y[0] + list_x[0] * list_y[0]
        
        for i in range(1,len(list_x)):
            carry= (list_x[i]+carry)*(list_y[i]+carry)-carry
        with tf.device(x_owner):
            xL = to_bool(carry.xL.inner_value)
            xL = tf.cast(xL, 'int64')
            xL = SharedTensorBase(inner_value=xL, module=2)
        with tf.device(y_owner):
            xR = to_bool(carry.xR.inner_value)
            xR = tf.cast(xR, 'int64')
            xR = SharedTensorBase(inner_value=xR, module=2)
        z = SharedPair(ownerL=x_owner, ownerR=y_owner, xL=xL, xR=xR, fixedpoint=0)
        z = z.reshape([-1])
        
        
        if z.shape[0] != np.prod(x.shape):
            z = z[0: np.prod(x.shape)]
        return z.reshape(x.shape)
    else:
        raise NotImplementedError


def leq(x: tf.Tensor, y: tf.Tensor, x_owner, y_owner) -> SharedPair:
    return geq(y, x, y_owner, x_owner)

def less(x: tf.Tensor, y: tf.Tensor, x_owner, y_owner) -> SharedPair:
    z = geq(x,y,x_owner,y_owner)
    return z.ones_like()-z

def greater(x: tf.Tensor, y: tf.Tensor, x_owner, y_owner) -> SharedPair:
    return less(y,x,y_owner,x_owner)


def compare_intfloat_intfloat(x_owner, x: tf.Tensor, y_owner, y: tf.Tensor, relation: str) -> SharedPairBase:
    x_owner=get_device(x_owner)
    y_owner=get_device(y_owner)
    if relation == "less":
        z = less(x, y, x_owner=x_owner, y_owner=y_owner)
    elif relation == "leq":
        z = leq(x, y, x_owner=x_owner, y_owner=y_owner)
    elif relation == "greater":
        z = greater(x, y, x_owner=x_owner, y_owner=y_owner)
    elif relation == "geq":
        z = geq(x, y, x_owner=x_owner, y_owner=y_owner)
    else:
        raise Exception("relation must be less, leq, greater or geq")
    return z


compare_int_int=compare_intfloat_intfloat

compare_float_float=compare_intfloat_intfloat



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : NN
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-09-11 14:29
   Description : description what the main function of this file
"""
from stensorflow.ml.nn.layers.layer import Layer
from stensorflow.basic.basic_class.pair import SharedVariablePair
from stensorflow.basic.basic_class.private import PrivateVariable
from stensorflow.ml.nn.layers.input import Input
import tensorflow as tf
import time
import numpy as np


class NN:
    """
    The class of neural network.
    """
    def __init__(self):
        self.layers = []

    def addLayer(self, ly: Layer):
        # 逐层添加
        # if fathers is not None:
        #     l.fathers = fathers
        for father in ly.fathers:

            if father not in self.layers:
                raise Exception("must add its fathers befor add it to network")
        self.layers += [ly]

    def compile(self):
        # 编译模型
        for ly in self.layers:
            for father in ly.fathers:
                if isinstance(father, Layer):
                    father.add_child(ly)
                else:
                    raise Exception("father must be a layer")
        l_last = self.layers[-1]
        assert isinstance(l_last, Layer)
        l_last.forward()

        for ly in self.layers:
            if isinstance(ly, Input):
                ly.backward()

    def get_train_sgd_op(self, learningRate, l2_regularization, momentum=0.0):
        """
        Construct a new Stochastic Gradient Descent
        """
        train_ops = []
        for ly in self.layers:
            if isinstance(ly, Layer):
                for i in range(len(ly.w)):
                    wi = ly.w[i]
                    assert isinstance(wi, SharedVariablePair) or isinstance(wi, PrivateVariable)
                    ploss_pwi = ly.ploss_pw[i] + l2_regularization * wi
                    if momentum > 0.0:
                        v = SharedVariablePair(ownerL=ploss_pwi.ownerL, ownerR=ploss_pwi.ownerR, shape=ploss_pwi.zeros_like())
                        v.load_from_numpy(np.zeros(shape=ploss_pwi.shape))
                        v_new = momentum * v + ploss_pwi
                        v_up_op = v.assign(v_new)
                        assign_op = wi.assign(wi - learningRate * v_new)
                        train_ops += [v_up_op, assign_op]
                    else:
                        assign_op = wi.assign(wi - learningRate * ploss_pwi)
                        train_ops += [assign_op]
        return tf.group(train_ops)

    def train_sgd(self, learning_rate, batch_num, l2_regularization, sess, momentum=0.0):
        learning_rate_list = None
        # 如果传入学习率list，转换为Tensor
        if isinstance(learning_rate, list):
            learning_rate_list = learning_rate
            self.learning_rate = tf.compat.v1.placeholder(dtype='float64', shape=[])

        else:
            self.learning_rate = learning_rate

        train_op = self.get_train_sgd_op(self.learning_rate, l2_regularization, momentum)
        sess.run(tf.compat.v1.global_variables_initializer())
        start_time = time.time()
        if learning_rate_list is not None:
            for i in range(batch_num):
                # 每次传入不同的学习率进行梯度下降
                print("batch ", i)
                sess.run(train_op, feed_dict={self.learning_rate: learning_rate_list[i]})
                if i%10==0:
                    print("time=", time.time()-start_time)
        else:
            for i in range(batch_num):
                print("batch ", i)
                sess.run(train_op)
                if i % 10 == 0:
                    print("time=", time.time() - start_time)


                # 输出某层结果的测试代码
                # print(self.layers[3])
                # tmp = sess.run([train_op, self.layers[3].y.to_tf_tensor("R")])
                # print(tmp[1])

    def cut_off(self):
        for ly in self.layers:
            assert isinstance(ly, Layer)
            ly.cut_off()

    def predict(self, x):
        raise NotImplementedError



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : Dense
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-09-11 15:57
   Description : description what the main function of this file
"""

from stensorflow.ml.nn.layers.layer import Layer
import numpy as np
from stensorflow.basic.basic_class.private import PrivateTensor, PrivateVariable
from stensorflow.basic.basic_class.pair import SharedVariablePair, SharedPair
from stensorflow.basic.basic_class.base import get_device
from typing import Union, List
from stensorflow.basic.protocol.bilinear_triangle import BiliinearTriangle
import pandas as pd
from stensorflow.exception.exception import StfNoneException, StfCondException


class Dense_bak(Layer):
    """
    Dense Layer
    """
    def __init__(self, output_dim, fathers, with_b=True):
        if fathers is None:
            raise StfNoneException("fathers")
        if fathers == []:
            raise StfCondException("fathers != []", "fathers == []")
        super(Dense, self).__init__(output_dim=output_dim, fathers=fathers)
        self.with_b = with_b
        for father in fathers:
            if not isinstance(father, Layer):
                raise Exception("father must be a layer")
            wi = SharedVariablePair(ownerL="L", ownerR="R", shape=[father.output_dim, output_dim])
            wi.load_from_numpy(
                np.random.normal(scale=1.0 / np.sqrt(father.output_dim + 1), size=[father.output_dim, output_dim]))
            self.w += [wi]

        if with_b:
            b = SharedVariablePair(ownerL="L", ownerR="R", shape=[output_dim])
            b.load_from_numpy(np.zeros([output_dim]))
            self.w += [b]

    def __str__(self):
        return "Dense Layer of output_dim={}".format(self.output_dim)

    def __repr__(self):
        return self.__str__()

    def func(self, w: List[SharedVariablePair], x: List[Union[PrivateTensor, SharedPair]]):
        if len(w) != len(x) + 1:
            raise Exception("must have len(w)==len(x)+1")

        y = x[0] @ w[0]
        y = y.dup_with_precision(x[0].fixedpoint)
        for i in range(1, len(x)):
            y = y + x[i] @ w[i]
        if self.with_b:
            y = y + self.w[len(x)]
        return y.dup_with_precision(new_fixedpoint=x[0].fixedpoint)

    def pull_back(self, w: List[SharedPair], x: List[Union[PrivateTensor, SharedPair]], y: SharedPair,
                  ploss_py: SharedPair) -> (List[SharedPair], List[SharedPair]):
        batch_size = x[0].shape[0]
        list_ploss_px = []
        ploss_pw = []
        for i in range(len(x)):
            ploss_pxi = ploss_py @ w[i].transpose()
            list_ploss_px += [ploss_pxi.dup_with_precision(x[0].fixedpoint)]

            ploss_pwi = x[i].transpose() @ ploss_py
            ploss_pwi = ploss_pwi.dup_with_precision(x[0].fixedpoint)
            ploss_pwi = ploss_pwi / batch_size
            ploss_pw += [ploss_pwi.dup_with_precision(x[0].fixedpoint)]

        ploss_px = dict(zip(self.fathers, list_ploss_px))

        if self.with_b:
            ploss_pb = ploss_py.reduce_sum(axis=[0]) / batch_size
            ploss_pw += [ploss_pb.dup_with_precision(x[0].fixedpoint)]

        return ploss_pw, ploss_px

    def save(self, save_file_machine, sess, path):
        j = 0
        for weight in self.w:
            weight = weight.to_tf_tensor(owner=save_file_machine)
            weight = sess.run(weight)
            weight = pd.DataFrame(data=weight)
            weight.to_csv(path + "_{}".format(j), header=False, index=False)
            j += 1

    def load(self, path):
        j = 0
        w = []
        for weight in self.w:
            assert isinstance(weight, SharedVariablePair) or isinstance(weight, PrivateVariable)
            value = pd.read_csv(path + "_{}".format(j), header=None, index_col=None)
            value = np.array(value)
            value = np.reshape(value, weight.shape)
            weight.load_from_numpy(value, const=True)
            w += [weight]
            j += 1
        self.w = w


class Dense(Layer):
    """
    Dense Layer
    """
    def __init__(self, output_dim, fathers, with_b=True):
        if fathers is None:
            raise StfNoneException("fathers")
        if fathers == []:
            raise StfCondException("fathers != []", "fathers == []")
        super(Dense, self).__init__(output_dim=output_dim, fathers=fathers)
        self.with_b = with_b
        self.bt_list = []
        for father in fathers:
            if not isinstance(father, Layer):
                raise Exception("father must be a layer")
            wi = SharedVariablePair(ownerL="L", ownerR="R", shape=[father.output_dim, output_dim])
            wi.load_from_numpy(
                np.random.normal(scale=1.0 / np.sqrt(father.output_dim + 1), size=[father.output_dim, output_dim]))
            self.w += [wi]
            f_xy = lambda a, b: a@b
            f_yz = lambda b, c: b@c.transpose()
            f_zx = lambda c, a: c.transpose()@a
            bt = BiliinearTriangle(f_xy, f_yz, f_zx) # x, w, plosspy^T
            self.bt_list.append(bt)

        if with_b:
            b = SharedVariablePair(ownerL="L", ownerR="R", shape=[output_dim])
            b.load_from_numpy(np.zeros([output_dim]))
            self.w += [b]

    def __str__(self):
        return "Dense Layer of output_dim={}".format(self.output_dim)

    def __repr__(self):
        return self.__str__()

    def func(self, w: List[SharedVariablePair], x: List[Union[PrivateTensor, SharedPair]]):
        if len(w) != len(x) + 1:
            raise Exception("must have len(w)==len(x)+1")

        # y = x[0] @ w[0]
        y = self.bt_list[0].compute_u(x[0], w[0])
        y = y.dup_with_precision(x[0].fixedpoint)
        for i in range(1, len(x)):
            # y = y + x[i] @ w[i]
            y = y + self.bt_list[i].compute_u(x[i], w[i])
        if self.with_b:
            y = y + self.w[len(x)]
        return y.dup_with_precision(new_fixedpoint=x[0].fixedpoint)

    def pull_back(self, w: List[SharedPair], x: List[Union[PrivateTensor, SharedPair]], y: SharedPair,
                  ploss_py: SharedPair) -> (List[SharedPair], List[SharedPair]):
        batch_size = x[0].shape[0]
        list_ploss_px = []
        ploss_pw = []
        for i in range(len(x)):
            # ploss_pxi = ploss_py @ w[i].transpose()
            # ploss_pwi = x[i].transpose() @ ploss_py
            ploss_pxi_t, ploss_pwi_t = self.bt_list[i].compute_vw(ploss_py)
            ploss_pxi = ploss_pxi_t.transpose()
            ploss_pwi = ploss_pwi_t.transpose()
            list_ploss_px += [ploss_pxi.dup_with_precision(x[0].fixedpoint)]
            ploss_pwi = ploss_pwi.dup_with_precision(x[0].fixedpoint)
            ploss_pwi = ploss_pwi / batch_size
            ploss_pw += [ploss_pwi.dup_with_precision(x[0].fixedpoint)]

        ploss_px = dict(zip(self.fathers, list_ploss_px))

        if self.with_b:
            ploss_pb = ploss_py.reduce_sum(axis=[0]) / batch_size
            ploss_pw += [ploss_pb.dup_with_precision(x[0].fixedpoint)]

        return ploss_pw, ploss_px

    def save(self, save_file_machine, sess, path):
        j = 0
        for weight in self.w:
            weight = weight.to_tf_tensor(owner=save_file_machine)
            weight = sess.run(weight)
            weight = pd.DataFrame(data=weight)
            weight.to_csv(path + "_{}".format(j), header=False, index=False)
            j += 1

    def load(self, path):
        j = 0
        w = []
        for weight in self.w:
            assert isinstance(weight, SharedVariablePair) or isinstance(weight, PrivateVariable)
            value = pd.read_csv(path + "_{}".format(j), header=None, index_col=None)
            value = np.array(value)
            value = np.reshape(value, weight.shape)
            weight.load_from_numpy(value, const=True)
            w += [weight]
            j += 1
        self.w = w







class Dense_Local(Layer):
    def __init__(self, output_dim, fathers, owner, with_b=True):
        super(Dense_Local, self).__init__(output_dim=output_dim, fathers=fathers)
        self.w = []
        self.owner = get_device(owner)
        self.with_b = with_b
        for father in fathers:
            if not isinstance(father, Layer):
                raise Exception("father must be a layer")
            wi = PrivateVariable(owner=self.owner)
            # wi.from_numpy(np.random.uniform(size=[father.output_dim, output_dim],
            #                                 low=-1.0 / np.sqrt(father.output_dim + 1),
            #                                 high=1.0 / np.sqrt(father.output_dim + 1)))

            wi.load_from_numpy(
                np.random.normal(scale=1.0 / np.sqrt(father.output_dim + 1), size=[father.output_dim, output_dim]))

            self.w += [wi]

        if with_b:
            b = PrivateVariable(owner=self.owner)
            b.load_from_numpy(np.zeros([output_dim]))
            self.w += [b]

    def func(self, w: List[PrivateTensor], x: List[Union[PrivateTensor, SharedPair]]):

        if (len(w) != len(x)) and (len(w) != len(x) + 1):
            raise Exception("must have len(w) == len(x) or len(w)==len(x)+1")

        y = PrivateTensor.from_PrivteTensorBase(x[0].to_private(owner=self.owner), op_map=x[0].op_map) @ w[0]

        y = y.dup_with_precision(x[0].fixedpoint)

        for i in range(1, len(x)):
            xi = x[i].to_private(owner=self.owner)

            y = y + xi @ w[i]

        if self.with_b:
            y = y + self.w[len(x)]

        return y.dup_with_precision(new_fixedpoint=x[0].fixedpoint)

    def pull_back(self, w: List[PrivateTensor], x: List[Union[PrivateTensor, SharedPair]], y: SharedPair,
                  ploss_py: SharedPair) -> (List[PrivateTensor], List[SharedPair]):
        batch_size = x[0].shape[0]
        list_ploss_px = []
        ploss_pw = []

        ploss_py = ploss_py.to_private(owner=self.owner)
        ploss_py = PrivateTensor.from_PrivteTensorBase(ploss_py)

        for i in range(len(x)):
            ploss_pxi = ploss_py @ w[i].transpose()
            list_ploss_px += [ploss_pxi.dup_with_precision(x[0].fixedpoint)]

            xi = x[i].to_private(owner=self.owner)
            xi = PrivateTensor.from_PrivteTensorBase(xi)
            ploss_pwi = xi.transpose() @ ploss_py
            ploss_pwi = ploss_pwi.dup_with_precision(x[0].fixedpoint)
            ploss_pwi = ploss_pwi / batch_size
            ploss_pw += [ploss_pwi.dup_with_precision(x[0].fixedpoint)]

        ploss_px = dict(zip(self.fathers, list_ploss_px))
        if self.with_b:
            ploss_pb = ploss_py.reduce_sum(axis=[0]) / batch_size
            ploss_pw += [ploss_pb.dup_with_precision(x[0].fixedpoint)]

        return ploss_pw, ploss_px

    def save(self, save_file_machine, sess, path):
        j = 0
        for weight in self.w:
            weight = weight.to_tf_tensor(owner=save_file_machine)
            weight = sess.run(weight)
            weight = pd.DataFrame(data=weight)
            weight.to_csv(path + "_{}".format(j), header=False, index=False)
            j += 1

    def load(self, path):
        j = 0
        w = []
        for weight in self.w:
            assert isinstance(weight, SharedVariablePair) or isinstance(weight, PrivateVariable)
            value = pd.read_csv(path + "_{}".format(j), header=None, index_col=None)
            value = np.array(value)
            value = np.reshape(value, weight.shape)
            weight.load_from_numpy(value, const=True)
            w += [weight]
            j += 1
        self.w = w



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : order.py
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-06-09 15:49
   Description : description what the main function of this file
"""

from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedPairBase, SharedTensorBase
import tensorflow as tf
from typing import Union
from stensorflow.basic.operator.logical import and_op, or_op, nxor_op, not_op
from stensorflow.basic.protocol.compare import greater_SharedTensorBase_SharedTensorBase


def to_bits(x: tf.Tensor, little_endian=True) -> tf.Tensor:
    if x.dtype != tf.int64:
        raise Exception("must have x.dtype==tf.int64")
    y = tf.bitcast(x, 'uint8')
    z = tf.bitwise.bitwise_and(tf.expand_dims(y, axis=-1), [1, 2, 4, 8, 16, 32, 64, 128])
    w = tf.bitwise.right_shift(z, [0, 1, 2, 3, 4, 5, 6, 7])
    new_shape = y.shape.as_list()
    new_shape[-1] = -1
    u = tf.reshape(w, new_shape)
    # if order == "decrease":
    if not little_endian:
        u = tf.reverse(u, axis=[-1])
    return tf.cast(u, 'int64')


def is_negative(x: SharedPairBase) -> SharedPairBase:
    with tf.device(x.ownerL):
        xL_bits = to_bits(x.xL.inner_value, little_endian=False)
        xL_bits_top, xL_bits_low = tf.split(xL_bits, [1, -1], axis=-1)
        xL_bits_top = SharedTensorBase(inner_value=xL_bits_top, module=2)
        xL_bits_low = SharedTensorBase(inner_value=xL_bits_low, module=2)
    with tf.device(x.ownerR):
        xR_bits = to_bits(x.xR.inner_value, little_endian=False)
        xR_bits_top, xR_bits_low = tf.split(xR_bits, [1, -1], axis=-1)
        xR_bits_top = SharedTensorBase(inner_value=xR_bits_top, module=2)
        xR_bits_low = SharedTensorBase(inner_value=xR_bits_low, module=2)

    carry = greater_SharedTensorBase_SharedTensorBase(xL_bits_low, -xR_bits_low + xR_bits_low.ones_like(), x.ownerL,
                                                      x.ownerR)
    x_bits_top = SharedPairBase(xL=xL_bits_top, xR=xR_bits_top, ownerL=x.ownerL, ownerR=x.ownerR, fixedpoint=0)
    x_bits_top = x_bits_top.squeeze(axis=[-1])
    result = carry + x_bits_top
    return result


def is_positive(x: SharedPairBase) -> SharedPairBase:
    return is_negative(x.zeros_like() - x)


def is_not_positive(x: SharedPairBase) -> SharedPairBase:
    y = is_positive(x)
    return y.ones_like() - y


def is_not_negative(x: SharedPairBase) -> SharedPairBase:
    y = is_negative(x)
    return y.ones_like() - y


def less(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase, int, float]) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    """

    :param x:
    :param y:
    :return:  1 if x<y else 0
    """
    if isinstance(x, PrivateTensorBase) and isinstance(y, PrivateTensorBase):
        if x.owner == y.owner:
            return PrivateTensorBase.__lt__(x, y)
        else:
            return NotImplemented
    elif isinstance(x, PrivateTensorBase) and isinstance(y, (int, float)):
        inner_value = tf.cast(x.to_tf_tensor(dtype='int64') < y, 'int64')
        return PrivateTensorBase(owner=x.owner, fixedpoint=0, inner_value=inner_value, module=2)
    elif isinstance(x, SharedPairBase) and isinstance(y, (int, float)):
        is_negative_x = is_negative(x)
        if y == 0:
            return is_negative_x
        same_sym_flag = is_negative_x if y < 0 else not_op(is_negative_x)
        same_less = and_op(same_sym_flag, is_negative(x - y))
        diff_less = is_negative_x if y >= 0 else same_sym_flag.zeros_like()
        return or_op(same_less, diff_less)
    elif isinstance(x, SharedPairBase) and isinstance(y, SharedPairBase):
        is_negative_x = is_negative(x)
        is_negative_y = is_negative(y)
        same_sym_flag = nxor_op(is_negative_x, is_negative_y)
        same_less = and_op(same_sym_flag, is_negative(x - y))
        diff_less = and_op(is_negative_x, not_op(is_negative_y))
        return or_op(same_less, diff_less)
    else:
        pass


def leq(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase]) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    """

    :param x:
    :param y:
    :return:  1 if x<=y else 0
    """
    y_less_x = less(y, x)
    return not_op(y_less_x)


def greater(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase]) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    """

    :param x:
    :param y:
    :return:  1 if x>y else 0
    """
    return less(y, x)


def geq(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase]) -> \
        Union[PrivateTensorBase, SharedPairBase]:
    """

    :param x:
    :param y:
    :return:  1 if x>=y else 0
    """
    x_less_y = less(x, y)
    return not_op(x_less_y)



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : global
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-05-14 14:30
   Description : description what the main function of this file
"""

import tensorflow as tf
import json
import platform
import os


class StfConfig:
    workerL_hosts = None
    workerR_hosts = None
    RS_hosts = None
    workerL = None
    workerR = None
    RS = None
    serverL = None
    serverR = None
    serverRS = None
    default_fixed_point = 14
    prf_flag = True
    compress_flag = True
    target = None
    train_file_onL = None
    train_file_onR = None
    pred_file_onL = None
    pred_file_onR = None
    predict_to_file = None
    custom_ops_path = None
    random_module = None
    conv_module = None
    pool_module = None
    homo_module = None
    upper_bd_int64 = (1 << 63) - 1
    lower_bd_int64 = -(1 << 63)
    to_str_precision = 6
    stf_home = None
    stf_home_workerL = None
    stf_home_workerR = None
    stf_home_RS = None
    coll_name_vars_random = "var_for_random"  # collection name for vars for random
    coll_name_vars_homo = "var_for_homo"  # collection name for vars for homo
    drelu = "log"
    invert_iter_num = 32
    softmax_iter_num = 32
    parties = 3
    pre_produce_flag = None
    offline_model = None
    pre_produce_list = []
    offline_triple_multiplex = 10000

    @classmethod
    def load_config(cls, config_file, job_name=None):
        with open(config_file) as f:
            config_dict = json.load(f)
        StfConfig.parties = config_dict.get("parties")
        if StfConfig.parties is None:
            StfConfig.parties = 3
        hosts_dict = config_dict.get("hosts")
        workerL_ip_port = hosts_dict.get("workerL")
        workerR_ip_port = hosts_dict.get("workerR")
        if StfConfig.parties == 3:
            RS_ip_port = hosts_dict.get("RS")


        StfConfig.workerL_hosts = workerL_ip_port.split(",")
        StfConfig.workerR_hosts = workerR_ip_port.split(",")
        if StfConfig.parties == 3:
            StfConfig.RS_hosts = RS_ip_port.split(",")

        workerL = map(lambda task_id: tf.DeviceSpec(job="workerL", task=task_id), range(len(StfConfig.workerL_hosts)))
        StfConfig.workerL = list(workerL)
        workerR = map(lambda task_id: tf.DeviceSpec(job="workerR", task=task_id), range(len(StfConfig.workerR_hosts)))
        StfConfig.workerR = list(workerR)
        if StfConfig.parties == 3:
            RS = map(lambda task_id: tf.DeviceSpec(job="RS", task=task_id), range(len(StfConfig.RS_hosts)))
            StfConfig.RS = list(RS)

        if StfConfig.parties == 2:
            StfConfig.pre_produce_flag = config_dict.get("pre_produce_flag")
            StfConfig.offline_model = config_dict.get("offline_model")



        if job_name=="workerL":
            sess_worker = "workerL"
        else:
            sess_worker = "workerR"

        if sess_worker == "workerL":
            sess_ip_port = workerL_ip_port
        elif sess_worker == "workerR":
            sess_ip_port = workerR_ip_port
        else:
            raise Exception("must have sess_worker==workerL or sess_worker==workerR "
                            "but sess_worker={}".format(sess_worker))
        StfConfig.target = "grpc://" + sess_ip_port

        StfConfig.prf_flag = config_dict.get("prf_flag")
        StfConfig.compress_flag = config_dict.get("compress_flag")
        StfConfig.default_fixed_point = config_dict.get("default_fixed_point")
        if job_name is None:
            stf_home = os.environ.get('stf_home')
            if stf_home is None:
                stf_home = config_dict.get("stf_home")
            StfConfig.stf_home = stf_home
            StfConfig.stf_home_workerL = stf_home
            StfConfig.stf_home_workerR = stf_home
            if StfConfig.parties == 3:
                StfConfig.stf_home_RS = stf_home
        else:
            StfConfig.stf_home_workerL = config_dict.get("stf_home_workerL")
            StfConfig.stf_home_workerR = config_dict.get("stf_home_workerR")
            if StfConfig.parties == 3:
                StfConfig.stf_home_RS = config_dict.get("stf_home_RS")
            if job_name == "workerL":
                StfConfig.stf_home = StfConfig.stf_home_workerL
            elif job_name == "workerR":
                StfConfig.stf_home = StfConfig.stf_home_workerR
            elif job_name == "RS":
                StfConfig.stf_home = StfConfig.stf_home_RS
            else:
                raise Exception("must have job_name in {'workerL', 'workerR', 'rs'}.")

        # -------ml config ---------------------
        ml_config = config_dict.get("ml")
        if ml_config is not None:
            train_cfg = ml_config.get("dataset_train")
            pred_cfg = ml_config.get("dataset_predict")
            StfConfig.train_file_onL = os.path.join(StfConfig.stf_home_workerL, train_cfg.get("L"))
            StfConfig.train_file_onR = os.path.join(StfConfig.stf_home_workerR, train_cfg.get("R"))
            StfConfig.pred_file_onL = os.path.join(StfConfig.stf_home_workerL, pred_cfg.get("L"))
            StfConfig.pred_file_onR = os.path.join(StfConfig.stf_home_workerR, pred_cfg.get("R"))
            if sess_worker == "workerL":
                StfConfig.predict_to_file = os.path.join(StfConfig.stf_home_workerL, ml_config.get("predict_to_file"))
            elif sess_worker == "workerR":
                StfConfig.predict_to_file = os.path.join(StfConfig.stf_home_workerR, ml_config.get("predict_to_file"))
            else:
                raise Exception("must have sess_worker==workerL or sess_worker==workerR, "
                                "but sess_worker={}".format(sess_worker))

        #StfConfig.cops_path = os.path.join(StfConfig.stf_home, config_dict.get("cops_path"))

        StfConfig.cops_path = os.environ['HOME'] + '/stf_cops'
        operating_system_platform = platform.system()

        if operating_system_platform == "Linux":
            StfConfig.random_module = tf.load_op_library(
                os.path.join(StfConfig.cops_path, '_stf_random_linux.so'))
            StfConfig.pool_module = tf.load_op_library(
                os.path.join(StfConfig.cops_path, '_stf_int64pooling_linux.so'))
            StfConfig.conv_module = tf.load_op_library(
                os.path.join(StfConfig.cops_path, '_stf_int64conv2d_linux.so'))
            StfConfig.homo_module = tf.load_op_library(
                os.path.join(StfConfig.cops_path, "_stf_homo_linux.so"))

        elif operating_system_platform == "Darwin":
            StfConfig.random_module = tf.load_op_library(
                os.path.join(StfConfig.cops_path, '_stf_random_macos.so'))
            StfConfig.pool_module = tf.load_op_library(
                os.path.join(StfConfig.cops_path, '_stf_int64pooling_macos.so'))
            StfConfig.conv_module = tf.load_op_library(
                os.path.join(StfConfig.cops_path, '_stf_int64conv2d_macos.so'))
            StfConfig.homo_module = tf.load_op_library(
                os.path.join(StfConfig.cops_path, "_stf_homo_macos.so"))
        else:
            raise Exception("only support Linux or macos.")

        # # ----------- protocols ------------
        protocols = config_dict.get("protocols")
        StfConfig.drelu = protocols.get("drelu")






# from stensorflow.basic.operator.arithmetic import *
# from stensorflow.basic.basic_class.private import *
# from stensorflow.basic.basic_class.pair import *
# from stensorflow.ml.logistic_regression import LogisticRegression
# from stensorflow.ml.logistic_regression2 import LogisticRegression2
#



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2021 All Rights Reserved.
   ------------------------------------------------------
   File Name : bitwise
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2021/2/22 下午4:11
   Description : description what the main function of this file
"""
from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedPairBase, SharedTensorBase, get_device
import tensorflow as tf
from stensorflow.global_var import StfConfig
from typing import Union


class SharedTensorBitwise:
    def __init__(self, inner_value, shape=None):
        """

        :param inner_value: A tf.Tensor of dtype = tf.int64
        :param shape:
        """
        self.inner_value = inner_value
        if shape is not None:
            self.shape = shape
        elif inner_value is not None:
            self.shape = inner_value.shape
        else:
            raise Exception("must have inner_value is not None or shape is not None")

    @classmethod
    def from_SharedTensor(cls, x: Union[SharedTensorBase, tf.Tensor]):
        if isinstance(x, SharedTensorBase):
            return SharedTensorBitwise(inner_value=x.inner_value)
        elif isinstance(x, tf.Tensor):
            return SharedTensorBitwise(inner_value=x)

    def to_SharedTensor(self):
        return SharedTensorBase(inner_value=self.inner_value)

    def __repr__(self) -> str:
        return 'SharedTensorBitwise(inner_value={})'.format(self.inner_value)

    def __add__(self, other):
        inner_value = tf.bitwise.bitwise_xor(self.inner_value, other.inner_value)
        return SharedTensorBitwise(inner_value=inner_value)

    def __sub__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        inner_value = tf.bitwise.bitwise_and(self.inner_value, other.inner_value)
        return SharedTensorBitwise(inner_value=inner_value)

    def random_uniform_adjoint(self, seed=None):
        x = self.to_SharedTensor()
        y = x.random_uniform_adjoint(seed=seed)
        return self.from_SharedTensor(y)

    def to_PrivateTensorBitwise(self, owner):
        with tf.device(get_device(owner)):
            return PrivateTensorBitwise(owner=owner, inner_value=self.inner_value)


class PrivateTensorBitwise:
    def __init__(self, owner, inner_value: SharedTensorBitwise, shape=None):
        """

        :param owner:  str: "L" or "R" or a python.framework.device_spec.DeviceSpecV2 object
        :param inner_value: tf.int64
        :param shape:
        """
        self.owner = get_device(owner)
        with tf.device(get_device(owner)):
            self.inner_value = inner_value
        if shape is not None:
            self.shape = shape
        elif inner_value is not None:
            self.shape = inner_value.shape
        else:
            raise Exception("must have inner_value is not None or shape is not None")

    @classmethod
    def from_PrivateTensorBase(cls, x: PrivateTensorBase):
        with tf.device(x.owner):
            inner_value = SharedTensorBitwise.from_SharedTensor(x.inner_value)
        return PrivateTensorBitwise(owner=x.owner, inner_value=inner_value)

    def to_SharedTensorBitwise(self):
        return self.inner_value

    def __repr__(self) -> str:
        return 'PrivateTensorBitwise(owner={}, inner_value={})'.format(self.owner, self.inner_value)

    def __add__(self, other):
        if isinstance(other, PrivateTensorBitwise):
            if other.owner == self.owner:
                with tf.device(get_device(self.owner)):
                    inner_value = self.inner_value + other.inner_value
                return PrivateTensorBitwise(owner=self.owner, inner_value=inner_value)
            else:
                return SharedPairBitwise(ownerL=self.owner, ownerR=other.owner, xL=self.inner_value,
                                         xR=other.inner_value)
        elif isinstance(other, SharedPairBitwise):
            if other.ownerL == self.owner:
                with tf.device(get_device(self.owner)):
                    xL = self.inner_value + other.xL
                return SharedPairBitwise(ownerL=other.ownerL, ownerR=other.ownerR, xL=xL, xR=other.xR)
            elif other.ownerR == self.owner:
                with tf.device(get_device(self.owner)):
                    xR = self.inner_value + other.xR
                return SharedPairBitwise(ownerL=other.ownerL, ownerR=other.ownerR, xL=other.xL, xR=xR)
            else:
                raise Exception("must have other.ownerL == self.owner or other.ownerR == self.owner")
        else:
            raise Exception("must have isinstance(other, PrivateTensorBitwise) or isinstance(other, SharedPairBitwise)")

    def random_uniform_adjoint(self, seed=None):
        x = self.to_SharedTensorBitwise()
        y = x.random_uniform_adjoint(seed=seed)
        return y

    def __sub__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        if isinstance(other, PrivateTensorBitwise):
            if other.owner == self.owner:
                with tf.device(get_device(self.owner)):
                    inner_value = tf.bitwise.bitwise_and(self.inner_value, other.inner_value)
                    return PrivateTensorBitwise(owner=self.owner, inner_value=inner_value)
            else:
                with tf.device(StfConfig.RS[0]):
                    self_adjoint = self.random_uniform_adjoint()
                    other_adjoint = other.random_uniform_adjoint()
                    z = self_adjoint * other_adjoint
                    zL = z.random_uniform_adjoint()
                    zR = z - zL
                with tf.device(self.owner):
                    dself = self.inner_value - self_adjoint
                with tf.device(other.owner):
                    dother = other.inner_value - other_adjoint
                    uR = dself * other.inner_value + zR
                with tf.device(self.owner):
                    uL = self_adjoint * dother + zL
                return SharedPairBitwise(ownerL=self.owner, ownerR=other.owner, xL=uL, xR=uR)
        elif isinstance(other, SharedPairBitwise):
            if self.owner == other.ownerL or self.owner == other.ownerR:
                xyL = self * other.xL.to_PrivateTensorBitwise(owner=other.ownerL)
                xyR = self * other.xR.to_PrivateTensorBitwise(owner=other.ownerR)
                return xyL + xyR
            else:
                raise Exception("must have self.owner == other.ownerL or self.owner == other.ownerR")
        else:
            raise Exception("must have isinstance(other, PrivateTensorBitwise) or isinstance(other, SharedPairBitwise)")


class SharedPairBitwise:
    def __init__(self, ownerL, ownerR, xL: SharedTensorBitwise, xR: SharedTensorBitwise, shape=None):
        self.ownerL = get_device(ownerL)
        self.ownerR = get_device(ownerR)
        with tf.device(get_device(self.ownerL)):
            self.xL = xL
        with tf.device(get_device(self.ownerR)):
            self.xR = xR
        if shape is not None:
            self.shape = shape
        elif xL is not None:
            self.shape = xL.shape
        else:
            raise Exception("must have shape is not None or xL is not None")

    def __repr__(self) -> str:
        return 'SharedPairBitwise(ownerL={}, ownerR={}, xL={}, xR={})'.format(self.ownerL, self.ownerR, self.xL,
                                                                              self.xR)

    def reflex(self):
        return SharedPairBitwise(ownerL=self.ownerR, ownerR=self.ownerL, xL=self.xL, xR=self.xR)

    def to_tf_tensor(self):
        return (self.xL + self.xR).inner_value

    def to_SharedPair(self):
        return SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR, xL=self.xL.to_SharedTensor(),
                              xR=self.xR.to_SharedTensor(), fixedpoint=0)

    @classmethod
    def from_SharedPairBase(cls, x: SharedPairBase):
        return SharedPairBitwise(ownerL=x.ownerL, ownerR=x.ownerR, xL=x.xL, xR=x.xR)

    def __add__(self, other):
        if isinstance(other, PrivateTensorBitwise):
            return other.__add__(self)
        elif isinstance(other, SharedPairBitwise):
            if self.ownerL == other.ownerL and self.ownerR == other.ownerR:
                with tf.device(self.ownerL):
                    xL = self.xL + other.xL
                with tf.device(self.ownerR):
                    xR = self.xR + other.xR
                return SharedPairBitwise(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR)
            elif self.ownerL == other.ownerR and self.ownerR == other.ownerL:
                other = other.reflex()
                return self.__add__(other)
            else:
                raise Exception("must have {self.ownerL, self.ownerR} == {other.ownerL, other.ownerR}")
        else:
            raise Exception("must have isinstance(other, PrivateTensorBitwise) or isinstance(other, SharedPairBitwise)")

    def __mul__(self, other):
        if isinstance(other, PrivateTensorBitwise):
            return other.__mul__(self)
        elif isinstance(other, SharedPairBitwise):
            if self.ownerL == other.ownerL and self.ownerR == other.ownerR:
                with tf.device(StfConfig.RS[0]):
                    xL_adjoint = self.xL.random_uniform_adjoint()
                    xR_adjoint = self.xR.random_uniform_adjoint()
                    yL_adjoint = other.xL.random_uniform_adjoint()
                    yR_adjoint = other.xR.random_uniform_adjoint()
                    z_adjoint = (xL_adjoint + xR_adjoint) * (yL_adjoint + yR_adjoint)
                    zL_adjoint = z_adjoint.random_uniform_adjoint()
                    zR_adjoint = z_adjoint - zL_adjoint
                with tf.device(self.ownerL):
                    dxL = self.xL - xL_adjoint
                    dyL = other.xL - yL_adjoint

                with tf.device(self.ownerR):
                    dxR = self.xR - xR_adjoint
                    dyR = other.xR - yR_adjoint
                    dx = dxL + dxR
                    dy = dyL + dyR
                    zR = dx * other.xR + xR_adjoint * dy + zR_adjoint

                with tf.device(self.ownerL):
                    zL = dx * other.xL + xL_adjoint * dy + zL_adjoint
                return SharedPairBitwise(ownerL=self.ownerL, ownerR=self.ownerR, xL=zL, xR=zR)
            elif self.ownerL == other.ownerR and self.ownerR == other.ownerL:
                other = other.reflex()
                return self.__mul__(other)
        else:
            raise Exception("must have isinstance(other, PrivateTensorBitwise) or isinstance(other, SharedPairBitwise)")

    def __sub__(self, other):
        return self.__add__(other)
